# -*- coding: utf-8 -*-
import os
import re
import sys
from operator import itemgetter

from lumbermill.BaseThreadedModule import BaseThreadedModule
from lumbermill.utils.Decorators import ModuleDocstringParser, setInterval
from lumbermill.utils.misc import TimedFunctionManager


@ModuleDocstringParser
class RegexParser(BaseThreadedModule):
    """
    Parse a string by named regular expressions.

    If regex matches, fields in the data dictionary will be set as defined in the named regular expression.
    Additionally the field "lumbermill.event_type" will be set containing the name of the regex.
    In the example below this would be "httpd_access_log".

    It is also possible to define multiple regexes with the same name. This allows for different log patterns
    for the same log type, e.g. apache access logs and nginx access logs.

    source_field: Field to apply the regex to.
    mark_unmatched_as: Set <lumbermill.event_type> to this value if regex did not match.
    break_on_match: Stop applying regex patterns after first match.
    hot_rules_first: Apply regex patterns based on their hit count.

    Configuration template:

    - RegexParser:
       source_field:                    # <default: 'data'; type: string; is: optional>
       mark_unmatched_as:               # <default: 'Unknown'; type: string; is: optional>
       break_on_match:                  # <default: True; type: boolean; is: optional>
       hot_rules_first:                 # <default: True; type: boolean; is: optional>
       field_extraction_patterns:       # <type: list; is: required>
        - httpd_access_log: ['(?P<httpd_access_log>.*)', 're.MULTILINE | re.DOTALL', 'findall']
       receivers:
        - NextModule
    """

    module_type = "parser"
    """Set module type"""

    def configure(self, configuration):
        # Call parent configure method
        BaseThreadedModule.configure(self, configuration)
        # Set defaults
        supported_regex_match_types = ['search', 'findall']
        self.timed_func_handler = None
        self.source_field = self.getConfigurationValue('source_field')
        self.mark_unmatched_as = self.getConfigurationValue('mark_unmatched_as')
        self.break_on_match = self.getConfigurationValue('break_on_match')
        self.hot_rules_first = self.getConfigurationValue('hot_rules_first')
        self.event_types = []
        self.fieldextraction_regexpressions = []
        self.logstash_patterns = {}
        self.readLogstashPatterns()
        for regex_config in configuration['field_extraction_patterns']:
            event_type = regex_config.keys()[0]
            regex_pattern = regex_config[event_type]
            regex_options = 0
            regex_match_type = 'search'
            if isinstance(regex_pattern, list):
                i = iter(regex_pattern)
                # Pattern is the first entry
                regex_pattern = i.next()
                # Regex options the second
                try:
                    regex_options = eval(i.next())
                except:
                    etype, evalue, etb = sys.exc_info()
                    self.logger.error("RegEx error for options %s. Exception: %s, Error: %s." % (regex_options, etype, evalue))
                    self.lumbermill.shutDown()
                    return
                # Regex match type the third (optional)
                try:
                    regex_match_type = i.next()
                except:
                    pass
            # Make sure regex_match_type is valid
            # At the moment only search and findall are supported
            if regex_match_type not in supported_regex_match_types:
                self.logger.error("RegEx error for match type %s. Only %s are supported." % (regex_options, supported_regex_match_types))
                self.lumbermill.shutDown()
                return
            try:
                regex_pattern = self.replaceLogstashPatterns(regex_pattern)
                regex = re.compile(regex_pattern, regex_options)
            except:
                etype, evalue, etb = sys.exc_info()
                self.logger.error("RegEx error for %s pattern %s. Exception: %s, Error: %s." % (event_type, regex_pattern, etype, evalue))
                self.lumbermill.shutDown()
            self.fieldextraction_regexpressions.append({'event_type': event_type, 'pattern': regex, 'match_type': regex_match_type, 'hitcounter': 0})

    def initAfterFork(self):
        if self.hot_rules_first:
            resort_fieldextraction_regexpressions_func = self.getResortFieldextractionRegexpressionsFunc()
            self.timed_func_handler = TimedFunctionManager.startTimedFunction(resort_fieldextraction_regexpressions_func)
        BaseThreadedModule.initAfterFork(self)

    def getResortFieldextractionRegexpressionsFunc(self):
        @setInterval(10)
        def resortFieldextractionRegexpressions():
            """Resort the regular expression list, according to hitcount. Might speed up matching"""
            self.fieldextraction_regexpressions = sorted(self.fieldextraction_regexpressions, key=itemgetter('hitcounter'), reverse=True)
            for regex_data in self.fieldextraction_regexpressions:
                regex_data['hitcounter'] = 0
        return resortFieldextractionRegexpressions

    def readLogstashPatterns(self):
        path = "%s/../assets/grok_patterns" % os.path.dirname(os.path.realpath(__file__))
        for (dirpath, dirnames, filenames) in os.walk(path):
            for filename in filenames:
                lines = [line.strip() for line in open('%s%s%s' % (dirpath, os.sep, filename))]
                for line_no, line in enumerate(lines):
                    if line == "" or line.startswith('#'):
                        continue
                    try:
                        pattern_name, pattern = line.split(' ', 1)
                        self.logstash_patterns[pattern_name] = pattern
                    except:
                        etype, evalue, etb = sys.exc_info()
                        self.logger.warning("Could not read logstash pattern in file %s%s%s, line %s. Exception: %s, Error: %s." % (dirpath,  os.sep, filename, line_no+1, etype, evalue))

    def replaceLogstashPatterns(self, regex_pattern):
        #print regex_pattern
        pattern_name_re = re.compile('%\{(.*?)\}')
        for match in pattern_name_re.finditer(regex_pattern):
            for pattern_name in match.groups():
                pattern_identifier = False
                if ':' in pattern_name:
                    pattern_name, pattern_identifier = pattern_name.split(':')
                try:
                    logstash_pattern = self.replaceLogstashPatterns(self.logstash_patterns[pattern_name])
                    if not pattern_identifier:
                        regex_pattern = regex_pattern.replace('%%{%s}' % pattern_name, logstash_pattern)
                    else:
                        regex_pattern = regex_pattern.replace('%%{%s:%s}' % (pattern_name, pattern_identifier), '(?P<%s>%s)' % (pattern_identifier, logstash_pattern))
                except KeyError:
                    self.logger.warning("Could not parse logstash pattern %s. Pattern name not found in pattern files." % (pattern_name))
                    continue
        return regex_pattern

    def handleEvent(self, event):
        """
        When an event type was successfully detected, extract the fields with to corresponding regex pattern.
        """
        if self.source_field not in event:
            yield event
            return
        if not isinstance(event[self.source_field], basestring):
            self.logger.warning("Data in event[%s] not of type string. Skipping." % self.source_field)
            yield event
            return
        string_to_match = event[self.source_field]
        matches_dict = False
        for regex_data in self.fieldextraction_regexpressions:
            event_type = regex_data['event_type']
            matches_dict = {}
            if regex_data['match_type'] == 'search':
                matches = regex_data['pattern'].search(string_to_match)
                if matches:
                    matches_dict = matches.groupdict()
            elif regex_data['match_type'] == 'findall':
                for match in regex_data['pattern'].finditer(string_to_match):
                    for key, value in match.groupdict().items():
                        try:
                            matches_dict[key].append(value)
                        except:
                            matches_dict[key] = [value]
            if matches_dict:
                event.update(matches_dict)
                event['lumbermill']['event_type'] = event_type
                if self.hot_rules_first:
                    regex_data['hitcounter'] += 1
                if(self.break_on_match):
                    break
        if not matches_dict:
            event['lumbermill']['event_type'] = self.mark_unmatched_as
        yield event


# -*- coding: utf-8 -*-
import hashlib
import re
import sys

from lumbermill.BaseThreadedModule import BaseThreadedModule
from lumbermill.utils.Decorators import ModuleDocstringParser


@ModuleDocstringParser
class ModifyFields(BaseThreadedModule):
    """
    Simple module to insert/delete/change field values.

    Configuration templates:

    # Keep all fields listed in source_fields, discard all others.
    - ModifyFields:
       action: keep                     # <type: string; is: required>
       source_fields:                   # <type: list; is: required>
       receivers:
        - NextModule

    # Discard all fields listed in source_fields.
    - ModifyFields:
       action: delete                   # <type: string; is: required>
       source_fields:                   # <type: list; is: required>
       receivers:
        - NextModule

    # Concat all fields listed in source_fields.
    - ModifyFields:
       action: concat                   # <type: string; is: required>
       source_fields:                   # <type: list; is: required>
       target_field:                    # <type: string; is: required>
       receivers:
        - NextModule

    # Insert a new field with "target_field" name and "value" as new value.
    - ModifyFields:
       action: insert                   # <type: string; is: required>
       target_field:                    # <type: string; is: required>
       value:                           # <type: string; is: required>
       receivers:
        - NextModule

    # Replace field values matching string "old" in data dictionary with "new".
    - ModifyFields:
       action: string_replace           # <type: string; is: required>
       source_field:                    # <type: string; is: required>
       old:                             # <type: string; is: required>
       new:                             # <type: string; is: required>
       max:                             # <default: -1; type: integer; is: optional>
       receivers:
        - NextModule

    # Replace field values in data dictionary with self.getConfigurationValue['with'].
    - ModifyFields:
       action: replace                  # <type: string; is: required>
       source_field:                    # <type: string; is: required>
       regex: ['<[^>]*>', 're.MULTILINE | re.DOTALL'] # <type: list; is: required>
       with:                            # <type: string; is: required>
       receivers:
        - NextModule

    # Rename a field.
    - ModifyFields:
       action: rename                   # <type: string; is: required>
       source_field:                    # <type: string; is: required>
       target_field:                    # <type: string; is: required>
       receivers:
        - NextModule

    # Rename a field by regex.
    - ModifyFields:
       action: rename_regex             # <type: string; is: required>
       regex:                           # <type: string; is: required>
       source_field:                    # <default: None; type: None||string; is: optional>
       target_field_pattern:            # <type: string; is: required>
       recursive:                       # <default: True; type: boolean; is: optional>
       receivers:
        - NextModule

    # Rename a field by replace.
    - ModifyFields:
       action: rename_replace           # <type: string; is: required>
       old:                             # <type: string; is: required>
       new:                             # <type: string; is: required>
       source_field:                    # <default: None; type: None||string; is: optional>
       recursive:                       # <default: True; type: boolean; is: optional>
       receivers:
        - NextModule

    # Map a field value.
    - ModifyFields:
       action: map                      # <type: string; is: required>
       source_field:                    # <type: string; is: required>
       map:                             # <type: dictionary; is: required>
       target_field:                    # <default: "$(source_field)_mapped"; type: string; is: optional>
       keep_unmappable:                 # <default: False; type: boolean; is: optional>
       receivers:
        - NextModule

    # Split source field to target fields based on key value pairs.
    - ModifyFields:
       action: key_value                # <type: string; is: required>
       line_separator:                  # <type: string; is: required>
       kv_separator:                    # <type: string; is: required>
       source_field:                    # <type: list; is: required>
       target_field:                    # <default: None; type: None||string; is: optional>
       prefix:                          # <default: None; type: None||string; is: optional>
       receivers:
        - NextModule

    # Split source field to target fields based on key value pairs using regex.
    - ModifyFields:
       action: key_value_regex          # <type: string; is: required>
       regex:                           # <type: string; is: required>
       source_field:                    # <type: list; is: required>
       target_field:                    # <default: None; type: None||string; is: optional>
       prefix:                          # <default: None; type: None||string; is: optional>
       receivers:
        - NextModule

    # Split source field to array at separator.
    - ModifyFields:
       action: split                    # <type: string; is: required>
       separator:                       # <type: string; is: required>
       source_field:                    # <type: list; is: required>
       target_field:                    # <default: None; type: None||string; is: optional>
       receivers:
        - NextModule

    # Merge source fields to target field as list.
    - ModifyFields:
       action: merge                    # <type: string; is: required>
       target_field:                    # <type: string; is: reuired>
       source_fields:                   # <type: list; is: required>
       receivers:
        - NextModule

    # Merge source field to target field as string.
    - ModifyFields:
       action: join                     # <type: string; is: required>
       source_field:                    # <type: string; is: required>
       target_field:                    # <type: string; is: required>
       separator:                       # <default: ","; type: string; is: optional>
       receivers:
        - NextModule

    # Cast field values to integer.
    - ModifyFields:
       action: cast_to_int              # <type: string; is: required>
       source_fields:                   # <type: list; is: required>
       receivers:
        - NextModule

    # Cast field values to float.
    - ModifyFields:
       action: cast_to_float            # <type: string; is: required>
       source_fields:                   # <type: list; is: required>
       receivers:
        - NextModule

    # Cast field values to string.
    - ModifyFields:
       action: cast_to_str              # <type: string; is: required>
       source_fields:                   # <type: list; is: required>
       receivers:
        - NextModule

    # Cast field values to boolean.
    - ModifyFields:
       action: cast_to_bool             # <type: string; is: required>
       source_fields:                   # <type: list; is: required>
       receivers:
        - NextModule

    # Create a hash from a field value.
    # If target_fields is provided, it should have the same length as source_fields.
    # If target_fields is not provided, source_fields will be replaced with the hashed value.
    # Hash algorithm can be any of the in hashlib supported algorithms.
    - ModifyFields:
       action: hash                     # <type: string; is: required>
       algorithm: sha1                  # <default: "md5"; type: string; is: optional;>
       salt:                            # <default: None; type: None||string; is: optional;>
       source_fields:                   # <type: list; is: required>
       target_fields:                   # <default: []; type: list; is: optional>
       receivers:
        - NextModule
"""

    module_type = "modifier"
    """Set module type"""

    def configure(self, configuration):
        # Call parent configure method
        BaseThreadedModule.configure(self, configuration)
        # Set defaultså
        self.typecast_switch = { 'int': self.cast_to_int,
                                 'integer': self.cast_to_int,
                                 'float': self.cast_to_float,
                                 'str': self.cast_to_str,
                                 'string': self.cast_to_str,
                                 'bool': self.cast_to_bool,
                                 'boolean': self.cast_to_bool,
                                }
        self.action = configuration['action']
        # Precompile regex for replacement if defined
        if 'regex' in configuration:
            regex_pattern = configuration['regex']
            regex_options = 0
            if isinstance(regex_pattern, list):
                i = iter(regex_pattern)
                # Pattern is the first entry
                regex_pattern = i.next()
                # Regex options the second
                try:
                    regex_options = eval(i.next())
                except:
                    etype, evalue, etb = sys.exc_info()
                    self.logger.error("RegEx error for options %s. Exception: %s, Error: %s" % (regex_options, etype, evalue))
                    self.lumbermill.shutDown()
            try:
                self.regex = re.compile(regex_pattern, regex_options)
            except:
                etype, evalue, etb = sys.exc_info()
                self.logger.error("RegEx error for pattern %s. Exception: %s, Error: %s" % (regex_pattern, etype, evalue))
                self.lumbermill.shutDown()
        self.source_field = self.getConfigurationValue('source_field') if "source_field" in self.configuration_data else []
        self.source_fields = self.getConfigurationValue('source_fields') if "source_fields" in self.configuration_data else []
        self.target_field = self.getConfigurationValue('target_field') if "target_field" in self.configuration_data else []
        self.target_fields = self.getConfigurationValue('target_fields') if "target_fields" in self.configuration_data else []
        # Call action specific configure method.
        if "configure_%s_action" % self.action in dir(self):
            getattr(self, "configure_%s_action" % self.action)()
        # Get action specific method
        try:
            self.event_handler = getattr(self, "%s" % self.action)
        except AttributeError:
            etype, evalue, etb = sys.exc_info()
            self.logger.error("ModifyFields action called that does not exist: %s. Exception: %s, Error: %s" % (self.action, etype, evalue))
            self.lumbermill.shutDown()

    def configure_rename_replace_action(self):
        self.recursive = self.getConfigurationValue('recursive')
        self.old = self.getConfigurationValue('old')
        self.new = self.getConfigurationValue('new')

    def configure_rename_regex_action(self):
        self.recursive = self.getConfigurationValue('recursive')
        self.target_field_pattern = self.getConfigurationValue('target_field_pattern')

    def configure_split_action(self):
        self.separator = self.getConfigurationValue('separator')

    def configure_key_value_action(self):
        self.line_separator = self.getConfigurationValue('line_separator')
        self.kv_separator = self.getConfigurationValue('kv_separator')
        self.prefix = self.getConfigurationValue('prefix')

    def configure_key_value_regex_action(self):
        self.prefix = self.getConfigurationValue('prefix')

    def configure_join_action(self):
        self.separator = self.getConfigurationValue('separator')

    def configure_anonymize_action(self):
        self.configure_hash_action()

    def configure_hash_action(self):
        # Import murmur hashlib if configured.
        self.salt = self.getConfigurationValue('salt') if self.getConfigurationValue('salt') else ""
        self.algorithm = self.getConfigurationValue('algorithm')
        if self.algorithm == "murmur":
            try:
                import mmh3
                self.hash_func = mmh3.hash
            except ImportError:
                etype, evalue, etb = sys.exc_info()
                self.logger.error("Exception: %s, Error: %s" % (etype, evalue))
                self.lumbermill.shutDown()
        else:
            try:
                self.hashlib_func = getattr(hashlib, self.algorithm)
            except ImportError:
                etype, evalue, etb = sys.exc_info()
                self.logger.error("Exception: %s, Error: %s" % (etype, evalue))
                self.lumbermill.shutDown()
                return
            self.hash_func = self.hashlibFunc

    def getStartMessage(self):
        """
        Return the module name.
        """
        source = self.source_field if self.source_field else self.source_fields
        target = self.target_field if self.target_field else self.target_fields
        if not source:
            return "%s: %s" % (self.action, target)
        elif not target:
            return "%s: %s" % (self.action, source)
        else:
            return "%s: %s => %s" % (self.action, source, target)

    def hashlibFunc(self, string):
        return self.hashlib_func(string).hexdigest()

    def handleEvent(self, event):
        try:
            event = self.event_handler(event)
        except AttributeError:
            etype, evalue, etb = sys.exc_info()
            self.logger.error("ModifyFields action %s threw an error. Exception: %s, Error: %s" % (self.action, etype, evalue))
            self.lumbermill.shutDown()
        yield event

    def keep(self,event):
        """
        Field names not listed in self.configuration_data['source_fields'] will be deleted from data dictionary.

        @param event: dictionary
        @return: event: dictionary
        """
        fields_to_del = set(event).difference(self.source_fields)
        for field in fields_to_del:
            # Do not delete internal event information.
            if field == 'lumbermill':
                continue
            event.pop(field, None)
        return event

    def delete(self, event):
        """
        Field names listed in ['source_fields'] will be deleted from data dictionary.

        @todo: pypy seems to handle simple tight loops better than
        - first building a set from data dictionary and
        - then get common keys from ['source_fields'] and data via intersection

        e.g.:
        fields_to_check = self.getConfigurationValue('source_fields').intersection(set(data))

        Still, if the field set is a large one, the above approach could be faster.

        This problem affects this and some more methods in this class.
        Maybe the code can be altered to take this into account.

        @param event: dictionary
        @return: event: dictionary
        """
        for field in self.source_fields:
            event.pop(field, None)
        return event

    def insert(self, event):
        """
        Insert a new field with a given value.

        @param event: dictionary
        @return: event: dictionary
        """
        event[self.target_field] = self.getConfigurationValue('value', event)
        return event



    def concat(self, event):
        """
        Field names listed in ['source_fields'] will be concatenated to a new string.
        The result will be stored in ['target_field']

        @param event: dictionary
        @return: event: dictionary
        """
        concat_str = ""
        for field in self.source_fields:
            try:
                concat_str = "%s%s" % (concat_str,event[field])
            except KeyError:
                pass
        event[self.target_field] = concat_str
        return event

    def replace(self, event):
        """
        Field value in data dictionary will be replaced with ['with']

        @param event: dictionary
        @return: event: dictionary
        """
        try:
            event[self.source_field] = self.regex.sub(self.getConfigurationValue('with', event), event[self.source_field])
        except KeyError:
            pass
        return event

    def rename(self, event):
        """
        Field name ['from'] in data dictionary will be renamed to ['to']

        @param event: dictionary
        @return: event: dictionary
        """
        try:
            event[self.target_field] = event.pop(self.source_field)
        except KeyError:
            pass
        return event

    def rename_regex(self, event):
        if self.source_field:
            try:
                dict_to_scan = event[self.source_field]
            except KeyError:
                return event
        else:
            dict_to_scan = event
        self._rename_regex_recursive(dict_to_scan)
        return event

    def _rename_regex_recursive(self, dict_to_scan):
        fields_to_rename = {}
        for field_name, field_value in dict_to_scan.iteritems():
            new_field_name = self.regex.sub(self.target_field_pattern, field_name)
            if field_name != new_field_name:
                fields_to_rename[field_name] = new_field_name
            if self.recursive and isinstance(field_value, dict):
                self._rename_regex_recursive(field_value)
        for old_field_name, new_field_name in fields_to_rename.iteritems():
            dict_to_scan[new_field_name] = dict_to_scan.pop(old_field_name)

    def rename_replace(self, event):
        self.event = event
        if self.source_field:
            try:
                dict_to_scan = event[self.source_field]
            except KeyError:
                return event
        else:
            dict_to_scan = event
        self._rename_replace_recursive(dict_to_scan)
        return event

    def _rename_replace_recursive(self, dict_to_scan):
        fields_to_rename = {}
        for field_name, field_value in dict_to_scan.iteritems():
            new_field_name = field_name.replace(self.old, self.new)
            if field_name != new_field_name:
                fields_to_rename[field_name] = new_field_name
            if self.recursive and isinstance(field_value, dict):
                self._rename_replace_recursive(field_value)
        for old_field_name, new_field_name in fields_to_rename.iteritems():
            dict_to_scan[new_field_name] = dict_to_scan.pop(old_field_name)

    def string_replace(self, event):
        """
        Field value matching string in data dictionary will be replace with new.

        @param event: dictionary
        @return: event: dictionary
        """
        try:
            event[self.source_field] = event[self.source_field].replace(self.getConfigurationValue('old', event), self.getConfigurationValue('new', event), self.getConfigurationValue('max'))
        except KeyError:
            pass
        return event

    def map(self, event):
        """
        Field values in data dictionary will be mapped to ['with'][data[field]].
        By default, the target field is ${fieldname}_mapped and can be overwritten by config
        value "target_field"

        Useful e.g. to map http status codes to human readable status codes.
        @param event: dictionary
        @return: event: dictionary
        """
        target_field = self.target_field if self.target_field else "%s_mapped" % self.source_field
        try:
            event[target_field] = self.getConfigurationValue('map', event)[event[self.source_field]]
        except KeyError:
            if self.getConfigurationValue('keep_unmappable'):
                event[target_field] = event[self.source_field]
            else:
                pass
        return event

    def key_value(self, event):
        """
        Split source field to target fields based on key value pairs.

        - module: ModifyFields
          action: key_value                           # <type: string; is: required>
          kv_separator:                               # <type: string; is: required>
          line_separator:                             # <default: None; type: None||string; is: required>
          source_field:                               # <type: list; is: required>
          target_field:                               # <default: None; type: None||string; is: optional>
          prefix:                                     # <default: None; type: None||string; is: optional>
          receivers:
            - NextModule

        @param event: dictionary
        @return: event: dictionary
        """
        try:
            if self.line_separator:
                kv_dict = dict(kv.split(self.kv_separator) for kv in event[self.source_field].split(self.line_separator))
            else:
                kv_dict = event[self.source_field].split(self.kv_separator)
        except:
            return event
        if self.prefix:
            kv_dict = dict(map(lambda (key, value): ("%s%s" % (self.prefix, str(key)), value), kv_dict.items()))
        if self.target_field:
            event[self.target_field] = kv_dict
        else:
            event.update(kv_dict)
        return event

    def key_value_regex(self, event):
        """
        Split source field to target fields based on key value pairs.

        - module: ModifyFields
          action: key_value_regex                     # <type: string; is: required>
          regex:                                      # <type: string; is: required>
          source_field:                               # <type: list; is: required>
          target_field:                               # <default: None; type: None||string; is: optional>
          prefix:                                     # <default: None; type: None||string; is: optional>
          receivers:
            - NextModule

        # ([^=&?]+)[=]([^&=?]+)

        @param event: dictionary
        @return: event: dictionary
        """
        try:
            kv_dict = dict(re.findall(self.regex, event[self.source_field]))
        except:
            return event
        if self.prefix:
            kv_dict = dict(map(lambda (key, value): ("%s%s" % (self.prefix, str(key)), value), kv_dict.items()))
        if self.target_field:
            event[self.target_field] = kv_dict
        else:
            event.update(kv_dict)
        return event

    def split(self, event):
        """
        Split source field to array at separator.

        - module: ModifyFields
          action: split                               # <type: string; is: required>
          separator:                                  # <type: string; is: required>
          source_field:                               # <type: list; is: required>
          target_field:                               # <default: None; type: None||string; is: optional>
          receivers:
            - NextModule

        @param event: dictionary
        @return: event: dictionary
        """
        try:
            values = event[self.source_field].split(self.separator)
        except:
            return event
        target_field = self.target_field if self.target_field else self.source_field
        event[target_field] = values
        return event

    def merge(self, event):
        """
        Merge source fields to target field as list.

        @param event: dictionary
        @return: event: dictionary
        """
        merged_fields = []
        for field in self.source_fields:
            try:
                merged_fields.append(event[field])
            except:
                pass
        event[self.target_field] = merged_fields
        return event

    def join(self, event):
        """
        Join source field to target field as string.

        @param event: dictionary
        @return: event: dictionary
        """
        #event.update({self.getConfigurationValue('target_field'): separator.join(fields)})
        try:
            event[self.target_field] = self.separator.join(event[self.source_field])
        except:
            pass
        return event

    def cast(self, event):
        """
        Field values in data dictionary will be cast to datatype set in ['type']
        This is just an alias function for the direct call to the castTo{DataType} method.

        @param event: dictionary
        @return: event: dictionary
        """
        try:
            return self.typecast_switch[self.getConfigurationValue('type'), event](event)
        except:
            pass

    def cast_to_int(self, event):
        """
       ['source_fields'] values in data dictionary will be cast to integer.

        @param event: dictionary
        @return: event: dictionary
        """
        for field in self.source_fields:
            try:
                event[field] = int(float(event[field]))
            except ValueError:
                event[field] = 0
            except KeyError:
                pass
        return event

    def cast_to_float(self, event):
        """
        ['source_fields'] values in data dictionary will be cast to float.

        @param event: dictionary
        @return: event: dictionary
        """
        for field in self.source_fields:
            try:
                event[field] = float(event[field])
            except ValueError:
                event[field] = 0
            except KeyError:
                pass
        return event

    def cast_to_str(self, event):
        """
        ['source_fields'] values in data dictionary will be cast to string.

        @param event: dictionary
        @return: event: dictionary
        """
        for field in self.source_fields:
            try:
                event[field] = str(event[field])
            except ValueError:
                event[field] = ""
            except KeyError:
                pass
        return event

    def cast_to_bool(self, event):
        """
        ['source_fields'] values in data dictionary will be cast to boolean.

        @param event: dictionary
        @return: event: dictionary
        """
        for field in self.source_fields:
            try:
                event[field] = bool(event[field])
            except ValueError:
                event[field] = False
            except KeyError:
                pass
        return event

    def anonymize(self, data):
        """
        Alias function for hash.
        """
        return self.hash(data)

    def hash(self, event):
        """
        ['source_fields'] values in data dictionary will hashed with hash algorithm set in configuration.

        @param event: dictionary
        @return: event: dictionary
        """
        for idx, field in enumerate(self.source_fields):
            target_fieldname = field if not self.target_fields else self.target_fields[idx]
            try:
                event[target_fieldname] = self.hash_func("%s%s" % (self.salt, event[field]))
            except:
                pass
        return event


# -*- coding: utf-8 -*-
import signal
import sys

import yaml

from lumbermill.BaseModule import BaseModule
from lumbermill.utils.Decorators import ModuleDocstringParser, setInterval
from lumbermill.utils.misc import TimedFunctionManager


@ModuleDocstringParser
class PackConfiguration(BaseModule.BaseModule):
    """
    Synchronize configuration from leader to pack members.
    Any changes to the leaders configuration will be synced to all pack followers.

    Locally configured modules of pack members will not be overwritten by the leaders configuration.

    Module dependencies: ['Pack']

    pack: Name of the pack module. Defaults to the standard Pack module.
    ignore_modules: List of module names to exclude from sync process.
    interval: Time in seconds between checks if master config did change.

    Configuration template:

    - PackConfiguration:
       pack:                            # <default: 'Pack'; type: string; is: optional>
       ignore_modules:                  # <default: []; type: list; is: optional>
       interval: 10                     # <default: 60; type: integer; is: optional>
    """

    module_type = "stand_alone"
    """Set module type"""

    can_run_forked = False

    def configure(self, configuration):
        # Call parent configure method
        BaseModule.BaseModule.configure(self, configuration)
        #self.logger.setLevel(logging.DEBUG)
        # Get pack module instance.
        mod_info = self.lumbermill.getModuleInfoById(self.getConfigurationValue('pack'))
        if not mod_info:
            self.logger.error("Could not start cluster configuration module. Required pack module %s not found. Please check your configuration." % (self.getConfigurationValue('pack')))
            self.lumbermill.shutDown()
            return
        self.pack = mod_info['instances'][0]
        if self.pack.is_leader:
            self.update_config_func = self.getMasterConfigurationUpdateFunc()
            self.pack.addHandler(action='update_configuration_request', callback=self.leaderHandleUpdateConfigurationRequest)
        else:
            self.pack.addHandler(action='discovery_finish', callback=self.followerHandleDiscoveryFinishRequest)
            self.pack.addHandler(action='update_configuration_reply', callback=self.followerHandleUpdateConfigurationReply)
            self.pack.addHandler(action='push_updated_configuration', callback=self.followerHandleUpdateConfigurationReply)

    def syncConfigurationToPack(self, configuration):
        """
        Push configuration to all pack members.

        In contrast to followers asking for new configurations, the leader can push a new config to all followers
        via a call to this function.
        """
        message = self.pack.getDefaultMessageDict(action='push_updated_configuration', custom_dict={'configuration': configuration})
        self.pack.sendMessageToPack(message)

    def getMasterConfigurationUpdateFunc(self):
        @setInterval(self.getConfigurationValue('interval'))
        def checkForUpdatedMasterConfiguration():
            filtered_running_configuration = self.filterIgnoredModules(self.lumbermill.getConfiguration())
            if self.filtered_startup_config != filtered_running_configuration:
                self.filtered_startup_config = filtered_running_configuration
                self.syncConfigurationToPack(self.filtered_startup_config)
        return checkForUpdatedMasterConfiguration

    def followerHandleDiscoveryFinishRequest(self, sender, message):
        """
        Request configuration from pack leader.
        """
        message = self.pack.getDefaultMessageDict(action='update_configuration_request')
        self.logger.debug('Requested configuration from leader %s.' % self.pack.discovered_leader.getHostName())
        self.pack.sendMessageToPackLeader(message)

    def leaderHandleUpdateConfigurationRequest(self, sender, message):
        """
        Receive update configuration requests and handle them.
        """
        try:
            pack_follower = self.pack.pack_followers[sender[0]]
        except KeyError:
            return
        message = self.pack.getDefaultMessageDict(action='update_configuration_reply', custom_dict={'configuration': self.filtered_startup_config})
        self.logger.debug('Sending update configuration reply to %s' % pack_follower.getHostName())
        self.pack.sendMessageToPackFollower(pack_follower, message)

    def followerHandleUpdateConfigurationReply(self, sender, message):
        sender_ip = sender[0]
        if not self.pack.discovered_leader or sender_ip != self.pack.discovered_leader.getIp():
            return
        self.logger.debug('Got update configuration reply from %s.' % self.pack.discovered_leader.getHostName())
        leader_configuration = message['configuration']
        leader_configuration = self.filterIgnoredModules(leader_configuration)
        if leader_configuration != self.filtered_startup_config:
            self.logger.info("Got new pack configuration from %s." % (self.pack.discovered_leader.getHostName()))
            self.lumbermill.setConfiguration(leader_configuration, merge=True)
            try:
                with open(self.lumbermill.getConfigurationFilePath(), 'w') as outfile:
                    outfile.write(yaml.dump(self.lumbermill.getConfiguration(), default_flow_style=False))
                # Send signal to reload LumberMill.
                signal.alarm(1)
            except:
                etype, evalue, etb = sys.exc_info()
                self.logger.warning("Could not update configuration file %s. Exception: %s, Error: %s." % (self.lumbermill.getConfigurationFilePath(), etype, evalue))

    def filterIgnoredModules(self, configuration):
        filtered_configuration = []
        for idx, module_info in enumerate(configuration):
            if type(module_info) is dict:
                module_name = module_info.keys()[0]
            elif type(module_info) is str:
                module_name = module_info
            else:
                self.logger.error('Unknown module configuration. Module: %s.' % module_info)
                self.lumbermill.shutDown()
            # Never sync Pack modules.
            if module_name in ['Pack', 'PackConfiguration']:
                continue
            # Filter ignored modules.
            if module_name not in self.getConfigurationValue('ignore_modules'):
                filtered_configuration.append(module_info)
        return filtered_configuration

    def start(self):
        # Get currently running configuration.
        self.filtered_startup_config = self.filterIgnoredModules(self.lumbermill.getConfiguration())
        if self.pack.is_leader:
            TimedFunctionManager.startTimedFunction(self.update_config_func)




# -*- coding: utf-8 -*-
import abc
import logging
import os
import re
import sys
from functools import wraps

from constants import LOGLEVEL_STRING_TO_LOGLEVEL_INT
from utils.ConfigurationValidator import ConfigurationValidator
from utils.DynamicValues import parseDynamicValue, mapDynamicValue, GP_DYNAMIC_VAL_REGEX_WITH_TYPES


class BaseModule:
    """
    Base class for all lumbermill modules.

    If you happen to override one of the methods defined here, be sure to know what you
    are doing. You have been warned ;)

    Configuration template:

    - module: SomeModuleName
       id:                               # <default: ""; type: string; is: optional>
       filter:                           # <default: None; type: None||string; is: optional>
       add_fields:                       # <default: {}; type: dict; is: optional>
       delete_fields:                    # <default: []; type: list; is: optional>
       event_type:                       # <default: None; type: None||string; is: optional>
       log_level:                        # <default: 'info'; type: string; values: ['info', 'warn', 'error', 'critical', 'fatal', 'debug']; is: optional>
       ...
       receivers:
        - ModuleName
        - ModuleAlias
    """

    module_type = "generic"
    """ Set module type. """
    can_run_forked = True

    def __init__(self, lumbermill):
        self.logger = logging.getLogger(self.__class__.__name__)
        self.lumbermill = lumbermill
        self.receivers = {}
        self.configuration_data = {}
        self.input_filter = None
        self.output_filters = {}
        self.process_id = os.getpid()

    def configure(self, configuration=None):
        """
        Configure the module.
        This method will be called by the LumberMill main class after initializing the module
        and after the configure method of the module is called.
        The configuration parameter contains k:v pairs of the yaml configuration for this module.

        @param configuration: dictionary
        @return: void
        """
        if configuration:
            self.configuration_data.update(configuration)
        self.parseDynamicValuesInConfiguration()
        # Set log level.
        self.logger.setLevel(LOGLEVEL_STRING_TO_LOGLEVEL_INT[self.getConfigurationValue('log_level').lower()])
        # Set default actions.
        self.delete_fields = self.getConfigurationValue('delete_fields')
        self.add_fields = self.getConfigurationValue('add_fields')
        self.event_type = self.getConfigurationValue('event_type')
        # Set input filter.
        if self.getConfigurationValue('filter'):
            self.setInputFilter(self.getConfigurationValue('filter'))
        # Set output filter per receiver if configured.
        if 'receivers' in self.configuration_data:
            for receiver_config in self.getConfigurationValue('receivers'):
                if not isinstance(receiver_config, dict):
                    continue
                receiver_name, receiver_filter_config = iter(receiver_config.items()).next()
                self.addOutputFilter(receiver_name, receiver_filter_config['filter'])
        self.checkConfiguration()

    def parseDynamicValuesInConfiguration(self):
        """
        Replace the configuration notation for dynamic values with pythons notation.
        E.g:
        filter: $(lumbermill.source_module) == 'TcpServer'
        parsed:
        filter: %(lumbermill.source_module)s == 'TcpServer'
        """
        # Copy dict since we might change it during iteration.
        configuration_data_copy = self.configuration_data.copy()
        for key, value in configuration_data_copy.iteritems():
            self.configuration_data[key] = parseDynamicValue(value)

    def checkConfiguration(self):
        configuration_errors = ConfigurationValidator().validateModuleConfiguration(self)
        if configuration_errors:
            self.logger.error("Could not configure module %s. Problems: %s." % (self.__class__.__name__, configuration_errors))
            self.lumbermill.shutDown()

    def getStartMessage(self):
        """
        Return a start message for the module. Can be overwritten to customize the message, e.g. to include port a server is
        listening on.
        """
        return 'started'

    def getConfigurationValue(self, key, mapping_dict={}, use_strftime=False):
        """
        Get a configuration value. This method encapsulates the internal configuration dictionary and
        takes care of replacing dynamic variables of the pattern e.g. $(field_name) with the corresponding
        entries of the mapping dictionary. Most of the time, this will be the data dictionary.

        It will also take care to return a default value if the module doc string provided one.
        """
        # Test if requested key exists.
        #config_setting = self.configuration_data[key]
        try:
            config_setting = self.configuration_data[key]
        except KeyError:
            self.logger.warning("Could not find configuration setting for: %s." % key)
            self.lumbermill.shutDown()
        if not isinstance(config_setting, dict):
            self.logger.debug("Configuration for key: %s is incorrect." % key)
            return False
        if config_setting['contains_dynamic_value'] is False or not mapping_dict:
            return config_setting.get('value')
        return mapDynamicValue(config_setting.get('value'), mapping_dict, use_strftime)

    def addReceiver(self, receiver_name, receiver):
        if self.module_type != "output":
            self.receivers[receiver_name] = receiver

    def setInputFilter(self, filter_string):
        """
        Convert input filter to lamba function.
        E.g. original filter:
        filter: $(lumbermill.source_module) == 'TcpServer'
        converts to:
        lambda event : event.get('lumbermill.source_module', False) == 'TcpServer'
        """
        filter_string_tmp = re.sub('^if\s+', "", filter_string)
        filter_string_tmp = "lambda event : " + re.sub(r"%\((.*?)\)(-?\d*[-\.\*]?\d*[sdf]?)", r"event.get('\1', False)", filter_string_tmp)
        try:
            event_filter = eval(filter_string_tmp)
        except:
            etype, evalue, etb = sys.exc_info()
            self.logger.error("Failed to compile filter: %s. Exception: %s, Error: %s." % (filter_string, etype, evalue))
            self.lumbermill.shutDown()
        # Wrap default receiveEvent method with filtered one.
        self.wrapReceiveEventWithFilter(event_filter, filter_string)

    def addOutputFilter(self, receiver_name, filter_string):
        """
        Convert output filter to lamba function.
        E.g. original filter:
        filter: $(lumbermill.source_module) == 'TcpServer'
        converts to:
        lambda event : event.get('lumbermill.source_module', False) == 'TcpServer'
        """
        # Output filter strings are not automatically parsed by parseDynamicValuesInConfiguration. So we need to do this here.
        filter_string_tmp = GP_DYNAMIC_VAL_REGEX_WITH_TYPES.sub(r"event.get('\1', False)", filter_string)
        filter_string_tmp = re.sub('^if\s+', "", filter_string_tmp)
        filter_string_tmp = "lambda event : " + filter_string_tmp
        try:
            output_filter = eval(filter_string_tmp)
            self.output_filters[receiver_name] = output_filter
        except:
            etype, evalue, etb = sys.exc_info()
            print(filter_string_tmp)
            self.logger.error("Failed to compile filter: %s. Exception: %s, Error: %s." % (filter_string, etype, evalue))
            self.lumbermill.shutDown()

    def getFilteredReceivers(self, event):
        if not self.output_filters:
            return self.receivers
        filterd_receivers = {}
        for receiver_name, receiver in self.receivers.items():
            if receiver_name not in self.output_filters:
                filterd_receivers[receiver_name] = receiver
                continue
            try:
                matched = self.output_filters[receiver_name](event)
            except:
                etype, evalue, etb = sys.exc_info()
                self.logger.warning("Output filter for %s failed. Exception: %s, Error: %s." % (receiver_name, etype, evalue))
                continue
            # If the filter succeeds, the data will be send to the receiver. The filter needs the event variable to work correctly.
            if matched:
                filterd_receivers[receiver_name] = receiver
        return filterd_receivers

    def initAfterFork(self):
        """
        Call startInterval on receiver buffered queue.
        This is done here since the buffer uses a thread to flush buffer in
        given intervals. The thread will not survive a fork of the main process.
        So we need to start this after the fork was executed.
        """
        self.process_id = os.getpid()
        for receiver_name, receiver in self.receivers.items():
            if hasattr(receiver, 'put'):
                receiver.startInterval()

    def commonActions(self, event):
        #if not self.input_filter or self.input_filter_matched:
        # Add fields if configured.
        if self.add_fields:
            for field_name, field_value in mapDynamicValue(self.add_fields, event).items():
                try:
                    event[field_name] = field_value
                except KeyError:
                    self.logger.warning("KeyError: could not set %s: %s" % (field_name, field_value))
        # Delete fields if configured.
        for field in self.delete_fields:
            try:
                event.pop(field, None)
            except KeyError:
                pass
        if self.event_type:
            event['lumbermill']['event_type'] = mapDynamicValue(self.event_type, event)
        return event

    def sendEvent(self, event, apply_common_actions=True):
        receivers = self.receivers if not self.output_filters else self.getFilteredReceivers(event)
        if not receivers:
            return
        if(apply_common_actions):
            event = self.commonActions(event)
        if len(receivers) > 1:
            event_clone = event.copy()
        copy_event = False
        for receiver in receivers.values():
            self.logger.debug("Sending event from %s to %s" % (self, receiver))
            if hasattr(receiver, 'receiveEvent'):
                receiver.receiveEvent(event if not copy_event else event_clone.copy())
            else:
                receiver.put(event if not copy_event else event_clone.copy())
            copy_event = True

    def receiveEvent(self, event):
        for event in self.handleEvent(event):
            if event:
                self.sendEvent(event)

    def wrapReceiveEventWithFilter(self, event_filter, filter_string):
        wrapped_func = self.receiveEvent
        @wraps(wrapped_func)
        def receiveEventFiltered(event):
            try:
                if event_filter(event):
                    wrapped_func(event)
                else:
                    self.sendEvent(event, apply_common_actions=False)
            except:
                etype, evalue, etb = sys.exc_info()
                self.logger.warning("Filter <%s> failed. Exception: %s, Error: %s." % (filter_string, etype, evalue))
                # Pass event to next module.
                self.sendEvent(event, apply_common_actions=False)
        self.receiveEvent = receiveEventFiltered

    @abc.abstractmethod
    def handleEvent(self, event):
        """
        Process the event.

        @param event: dictionary
        """
        yield event

    def shutDown(self):
        self.alive = False


# -*- coding: utf-8 -*-
import tornado.web
import tornado.escape
import socket
from pygments import highlight
from pygments.lexers import YamlLexer
from pygments.formatters import HtmlFormatter

class BaseHandler(tornado.web.RequestHandler):
    @property
    def webserver_module(self):
        return self.settings['TornadoWebserver']

    def __get_current_user(self):
        user_json = self.get_secure_cookie("gambolputty_web")
        if not user_json: return None
        return tornado.escape.json_decode(user_json)

class MainHandler(BaseHandler):
    def get(self):
        nodes = []
        localnode = {'server_name': socket.gethostname(),
                     'server_type': 'StandAlone'}
        # Check if cluster module is available.
        cluster_info = self.webserver_module.gp.getModuleInfoById('Cluster')
        if cluster_info:
            cluster_module = cluster_info['instances'][0]
            localnode['server_type'] = "PackLeader" if cluster_module.leader else "PackMember"
            if cluster_module.leader:
                for pack_member in cluster_module.getPackMembers().itervalues():
                    nodes.append({'server_name': pack_member.getHostName(),
                                  'server_type': "PackLeader" if pack_member.leader else "PackMember"})
        nodes.insert(0, localnode)
        self.render("index.html",
                    page_title="LumberMill WebGui",
                    nodes=nodes,
                    webserver_module=self.webserver_module)

class ServerConfigurationAsText(BaseHandler):
    def get(self):
        config_as_html = highlight(self.webserver_module.gp.getRawConfiguration(), YamlLexer(), HtmlFormatter(noclasses=True))
        self.content_type = 'text/html'
        self.write(config_as_html)

class ConfigurationHandler(BaseHandler):
    def get(self):
        self.render("server_configuration.html",
                    page_title="Server configuration")


# -*- coding: utf-8 -*-
import socket
import sys

import zmq

import lumbermill.utils.DictUtils as DictUtils
from lumbermill.BaseThreadedModule import BaseThreadedModule
from lumbermill.utils.Decorators import ModuleDocstringParser


@ModuleDocstringParser
class Zmq(BaseThreadedModule):
    """
    Read events from a zeromq.


    mode: Whether to run a server or client.
    address: Address to connect to. Pattern: hostname:port. If mode is server, this sets the addresses to listen on.
    pattern: One of 'pull', 'sub'.
    hwm: Highwatermark for sending/receiving socket.

    Configuration template:

    - Zmq:
       mode:                            # <default: 'server'; type: string; values: ['server', 'client']; is: optional>
       address:                         # <default: '*:5570'; type: string; is: optional>
       pattern:                         # <default: 'pull'; type: string; values: ['pull', 'sub']; is: optional>
       topic:                           # <default: ''; type: string; is: optional>
       hwm:                             # <default: None; type: None||integer; is: optional>
       receivers:
        - NextModule
    """

    module_type = "input"
    """Set module type"""
    can_run_forked = False

    def configure(self, configuration):
        # Call parent configure method
        BaseThreadedModule.configure(self, configuration)
        self.topic = self.getConfigurationValue('topic')
        self.pattern = self.getConfigurationValue('pattern')
        self.context = zmq.Context()
        if self.pattern == 'pull':
            self.socket = self.context.socket(zmq.PULL)
        elif self.pattern == 'sub':
            self.socket = self.context.socket(zmq.SUB)
            self.socket.setsockopt(zmq.SUBSCRIBE, str(self.topic))
        if self.getConfigurationValue('hwm'):
            self.setReceiveHighWaterMark(self.getConfigurationValue('hwm'))
        server_addr, server_port = self.getServerAddress(self.getConfigurationValue('address'))
        if self.getConfigurationValue('mode') == 'server':
            self.bind(server_addr, server_port)
        else:
            self.connect(server_addr, server_port)

    def getServerAddress(self, server_name):
        # ZMQ does not like hostnames too much. Try to get the ip address for server name.
        server_name, server_port = server_name.split(":")
        try:
            server_addr = socket.gethostbyname(server_name)
        except socket.gaierror:
            server_addr = server_name
        return (server_addr, server_port)

    def setReceiveHighWaterMark(self, hwm):
        try:
            self.socket.setsockopt(zmq.RCVHWM, hwm)
        except:
            self.socket.setsockopt(zmq.HWM, hwm)

    def bind(self, server_addr, server_port):
        try:
            self.socket.bind('tcp://%s:%s' % (server_addr, server_port))
        except:
            etype, evalue, etb = sys.exc_info()
            self.logger.error("Could not bind to %s:%s. Exception: %s, Error: %s." % (server_addr, server_port, etype, evalue))
            self.lumbermill.shutDown()

    def connect(self, server_addr, server_port):
        try:
            self.socket.connect('tcp://%s:%s' % (server_addr, server_port))
        except:
            etype, evalue, etb = sys.exc_info()
            self.logger.error("Could not connect to zeromq at %s:%s. Exception: %s, Error: %s." % (server_addr, server_port, etype, evalue))
            self.lumbermill.shutDown()

    def getEventFromZmq(self):
        try:
            event = self.socket.recv()
            yield event
        except zmq.error.ContextTerminated:
            pass
        except:
            exc_type, exc_value, exc_tb = sys.exc_info()
            if exc_value in ['Interrupted system call', 'Socket operation on non-socket']:
                return
            self.logger.error("Could not read data from zeromq. Exception: %s, Error: %s." % (exc_type, exc_value))

    def run(self):
        if not self.receivers:
            self.logger.error("Shutting down module %s since no receivers are set." % (self.__class__.__name__))
            return
        while self.alive:
            for event in self.getEventFromZmq():
                event = DictUtils.getDefaultEventDict({"data": event}, caller_class_name="Zmq")
                if self.pattern == 'sub':
                    topic, event['data'] = event['data'].split(' ', 1)
                    event['topic'] = topic
                self.sendEvent(event)

    def shutDown(self):
        # Call parent shutDown method.
        BaseThreadedModule.shutDown(self)
        try:
            self.socket.close()
            self.context.term()
        except AttributeError:
            pass




# -*- coding: utf-8 -*-
import time

import lumbermill.utils.DictUtils as DictUtils
from lumbermill.BaseThreadedModule import BaseThreadedModule
from lumbermill.utils.Decorators import ModuleDocstringParser


@ModuleDocstringParser
class Spam(BaseThreadedModule):
    """
    Emits events as fast as possible.

    Use this module to load test LumberMill. Also nice for testing your regexes.

    The event field can either be a simple string. This string will be used to create a default lumbermill event dict.
    If you want to provide more custom fields, you can provide a dictionary containing at least a "data" field that
    should your raw event string.

    events: Send custom event data. For single events, use a string or a dict. If a string is provided, the contents will
            be put into the events data field.
            if a dict is provided, the event will be populated with the dict fields.
            For multiple events, provide a list of stings or dicts.
    sleep: Time to wait between sending events.
    events_count: Only send configured number of events. 0 means no limit.

    Configuration template:

    - Spam:
       event:                           # <default: ""; type: string||list||dict; is: optional>
       sleep:                           # <default: 0; type: int||float; is: optional>
       events_count:                    # <default: 0; type: int; is: optional>
       receivers:
        - NextModule
    """

    module_type = "input"
    """Set module type"""
    can_run_forked = True

    def configure(self, configuration):
        # Call parent configure method
        BaseThreadedModule.configure(self, configuration)
        self.events = self.getConfigurationValue("event")
        if not isinstance(self.events, list):
            self.events = [self.events]
        self.sleep = self.getConfigurationValue("sleep")
        self.max_events_count = self.getConfigurationValue("events_count")

    def initAfterFork(self):
        BaseThreadedModule.initAfterFork(self)
        # Calculate event count when running in multiple processes.
        if self.max_events_count == 0:
            return
        self.max_events_count = int(self.getConfigurationValue("events_count")/self.lumbermill.getWorkerCount())
        if self.lumbermill.is_master():
            remainder = self.getConfigurationValue("events_count") % self.lumbermill.getWorkerCount()
            self.max_events_count += remainder
        if self.max_events_count == 0:
            self.shutDown()

    def run(self):
        counter = 0
        while self.alive:
            for event_data in self.events:
                if isinstance(event_data, str):
                    event = DictUtils.getDefaultEventDict({'data': event_data}, caller_class_name=self.__class__.__name__)
                elif isinstance(event_data, dict):
                    event = DictUtils.getDefaultEventDict(event_data, caller_class_name=self.__class__.__name__) # self.getConfigurationValue("event")
                self.sendEvent(event)
                if self.sleep > 0:
                    time.sleep(self.sleep)
                if self.max_events_count == 0:
                    continue
                counter += 1
                if (counter - self.max_events_count == 0):
                    time.sleep(2)
                    self.alive = False
        self.lumbermill.shutDown()


# -*- coding: utf-8 -*-
import socket
import sys
import time

from lumbermill.BaseThreadedModule import BaseThreadedModule
from lumbermill.utils.Buffers import Buffer
from lumbermill.utils.Decorators import ModuleDocstringParser


@ModuleDocstringParser
class GraphiteSink(BaseThreadedModule):
    """
    Send metrics to graphite server.

    server: Graphite server to connect to.
    port: Port carbon-cache is listening on.
    formats: Format of messages to send to graphite, e.g.: ['lumbermill.stats.event_rate_$(interval)s $(event_rate)'].
    store_interval_in_secs: Send data to graphite in x seconds intervals.
    batch_size: Send data to graphite if event count is above, even if store_interval_in_secs is not reached.
    backlog_size: Send count of events waiting for transmission. Events above count will be dropped.

    Here a simple example to send http_status statistics to graphite:

    ...

    - Statistics:
       interval: 10
       fields: ['http_status']

    - GraphiteSink:
       filter: if $(field_name) == "http_status"
       server: 127.0.0.1
       batch_size: 1
       formats: ['lumbermill.stats.http_200_$(interval)s $(field_counts.200)',
                 'lumbermill.stats.http_400_$(interval)s $(field_counts.400)',
                 'lumbermill.stats.http_total_$(interval)s $(total_count)']

    ...

    Configuration template:

    - GraphiteSink:
       server:                          # <default: 'localhost'; type: string; is: optional>
       port:                            # <default: 2003; type: integer; is: optional>
       formats:                         # <type: list; is: required>
       store_interval_in_secs:          # <default: 5; type: integer; is: optional>
       batch_size:                      # <default: 50; type: integer; is: optional>
       backlog_size:                    # <default: 50; type: integer; is: optional>
    """

    module_type = "output"
    """Set module type"""

    def configure(self, configuration):
        # Call parent configure method
        BaseThreadedModule.configure(self, configuration)
        self.formats = self.getConfigurationValue('formats')
        self.connection_data = (self.getConfigurationValue('server'), self.getConfigurationValue('port'))
        self.connection = None

    def connect(self):
        # Connect to server
        connection = socket.socket()
        try:
            connection.connect(self.connection_data)
            return connection
        except:
            etype, evalue, etb = sys.exc_info()
            self.logger.error("Failed to connect to %s. Exception: %s, Error: %s." % (self.connection_data, etype, evalue))
            return False

    def getStartMessage(self):
        return "%s:%s. Max buffer size: %d" % (self.connection_data[0], self.connection_data[1], self.getConfigurationValue('backlog_size'))

    def initAfterFork(self):
        BaseThreadedModule.initAfterFork(self)
        self.buffer = Buffer(self.getConfigurationValue('batch_size'), self.storeData, self.getConfigurationValue('store_interval_in_secs'), maxsize=self.getConfigurationValue('backlog_size'))
        self.connection = self.connect()
        if not self.connection:
            self.lumbermill.shutDown()
            return
        BaseThreadedModule.initAfterFork(self)

    def handleEvent(self, event):
        for format in self.formats:
            mapped_data = self.mapDynamicValue(format, event)
            if mapped_data:
                self.buffer.append("%s %s" % (mapped_data, int(time.time())))
        yield None

    def storeData(self, events):
        for event in events:
            try:
                if not event.endswith("\n"):
                    event += "\n"
                self.connection.send(event)
                return True
            except:
                etype, evalue, etb = sys.exc_info()
                self.logger.error("Server communication error. Exception: %s, Error: %s." % (etype, evalue))
                tries = 0
                self.connection.close()
                self.connection = None
                while tries < 5 and not self.connection:
                    time.sleep(5)
                    self.logger.warning("Trying to reconnect to %s." % (self.connection_data))
                    # Try to reconnect.
                    self.connection = self.connect()
                    tries += 1
                if not self.connection:
                    self.logger.error("Reconnect failed. Shutting down.")
                    self.lumbermill.shutDown()
                else:
                    self.logger.info("Reconnection to %s successful." % (self.connection_data))

    def shutDown(self):
        try:
            self.connection.close()
        except:
            pass


#!/usr/bin/env python
# -*- coding: utf-8 -*-

# Copyright (c) 2014 Björn Puttmann

# Permission is hereby granted, free of charge, to any person obtaining
# a copy of this software and associated documentation files (the
# "Software"), to deal in the Software without restriction, including
# without limitation the rights to use, copy, modify, merge, publish,
# distribute, sublicense, and/or sell copies of the Software, and to
# permit persons to whom the Software is furnished to do so, subject to
# the following conditions:

# The above copyright notice and this permission notice shall be
# included in all copies or substantial portions of the Software.

# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
# EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
# MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
# NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE
# LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION
# OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION
# WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
import sys

from setuptools import find_packages
from lumbermill import __version__, __app_name__, __author__, __email__, __url__

try:
    from setuptools import setup
    setup  # workaround for pyflakes issue #13
except ImportError:
    from distutils.core import setup

try:
    import __pypy__
    is_pypy = True
except ImportError:
    is_pypy = False

requirements = open('requirements/requirements.txt').readlines()
if is_pypy:
    requirements.extend(open('requirements/requirements-pypy.txt').readlines())

# Strip comments and newlines from requirements.
parsed_requirements = []
for req in requirements:
    req = req.strip()
    if not req or req.startswith('#'):
        continue
    parsed_requirements.append(req)

setup(
    name=__app_name__,
    version=__version__,
    author=__author__,
    author_email=__email__,
    packages=find_packages(exclude=['tests']),
    url=__url__,
    #data_files=[('conf', [os.path.join('conf', _) for _ in os.listdir('conf') if _.startswith('example-')]),
    #            ('scripts', [os.path.join('scripts', _) for _ in os.listdir('scripts') if _.startswith('spam')]),
    #            ('init.d', ['scripts/etc/init.d/lumbermill'])],
    include_package_data=True,
    install_requires=parsed_requirements,
    license='LICENSE',
    classifiers=[
        'Intended Audience :: System Administrators',
        'License :: OSI Approved :: Apache Software License',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: Implementation :: PyPy',
        'Topic :: System :: Logging',
    ],
    description='A logparser with module support.',
    long_description=open('README.rst').read() + '\n\n',
    entry_points={"console_scripts": ['lumbermill = lumbermill.LumberMill:main']},
    options={'build_scripts': {
                'executable': sys.executable},
    }
)


# -*- coding: utf-8 -*-
import copy
import gc
import random
import sys

import lumbermill.utils.DictUtils as DictUtils
from lumbermill.BaseThreadedModule import BaseThreadedModule
from lumbermill.utils.Decorators import ModuleDocstringParser, setInterval
from lumbermill.utils.misc import TimedFunctionManager


@ModuleDocstringParser
class EventBuffer(BaseThreadedModule):
    """
    Store received events in a persistent backend until the event was successfully handled.
    Events, that did not get handled correctly, will be requeued when LumberMill is restarted.

    At the moment only RedisStore is supported as backend.

    As a technical note: This module is based on pythons garbage collection. If an event is
    created, a copy of the event is stored in the persistence backend. If it gets garbage collected,
    the event will be deleted from the backend.
    When used, this module forces a garbage collection every <gc_interval> seconds.
    This approach seemed to be the fastest and simplest with a small drawback:
    IMPORTANT: It is not absolutely guaranteed, that an event will be collected, thus the event will
    not be deleted from the backend data. This can cause a limited amount of duplicate events being
    send to the sinks.
    With an elasticsearch sink, this should be no problem, as long as your document id
    stays the same for the same event data. This is also true for the default event_id.

    Configuration template:

    - EventBuffer:
       backend:                         # <default: 'RedisStore'; type: string; is: optional>
       gc_interval:                     # <default: 5; type: integer; is: optional>
       key_prefix:                      # <default: "lumbermill:eventbuffer"; type: string; is: optional>
       """

    module_type = "stand_alone"
    """Set module type"""
    can_run_forked = True

    def configure(self, configuration):
        # Call parent configure method
        BaseThreadedModule.configure(self, configuration)
        self.key_prefix = self.getConfigurationValue('key_prefix')
        self.key_buffer = {}
        self.flush_interval = self.getConfigurationValue('gc_interval')
        self.requeue_events_done = False
        backend_info = self.lumbermill.getModuleInfoById(self.getConfigurationValue('backend'))
        if not backend_info:
            self.logger.error("Could not find %s backend for persistant storage." % (self.getConfigurationValue('backend')))
            self.lumbermill.shutDown()
            return
        self.persistence_backend = backend_info['instances'][0]
        DictUtils.KeyDotNotationDict.persistence_backend = self.persistence_backend
        DictUtils.KeyDotNotationDict.key_prefix = self.key_prefix
        # Monkeypatch Utils.KeyDotNotationDict to add/delete event to persistence backend.
        def removeFromPersistenceBackendOnGarbageCollect(self):
            DictUtils.KeyDotNotationDict.___del___(self)
            # Only act if we hold an event.
            if "event_id" not in self.get("lumbermill", {}):
                return
            #print "Removing from backend"
            try:
                key = "%s:%s" % (DictUtils.KeyDotNotationDict.key_prefix, self['lumbermill']['event_id'])
                DictUtils.KeyDotNotationDict.persistence_backend.delete(key)
            except:
                pass
        DictUtils.KeyDotNotationDict.___del___ = DictUtils.KeyDotNotationDict.__del__
        DictUtils.KeyDotNotationDict.__del__ = removeFromPersistenceBackendOnGarbageCollect

        def addToPersistenceBackendOnInit(self, *args):
            DictUtils.KeyDotNotationDict.___init___(self, *args)
            # Only act if we hold an event.
            if "event_id" not in self.get("lumbermill", {}):
                return
            try:
                key = "%s:%s" % (DictUtils.KeyDotNotationDict.key_prefix, self['lumbermill']['event_id'])
                # Store a simple dict in backend, not a KeyDotNotationDict.
                # Also, store a copy, as the dict might get buffered in persistence_backend and we do not
                # want any changes made to new_dict to propagate to persistence_backend.
                DictUtils.KeyDotNotationDict.persistence_backend.set(key, dict.copy(self), False)
            except:
                etype, evalue, etb = sys.exc_info()
                self.logger.error("Could not store event in persistance backend. Exception: %s, Error: %s." % (etype, evalue))
                pass
            print "Added to backend"
        DictUtils.KeyDotNotationDict.___init___ = DictUtils.KeyDotNotationDict.__init__
        DictUtils.KeyDotNotationDict.__init__ = addToPersistenceBackendOnInit

        def returnSimpleKeyDotDictOnCopy(self):
            """
            When creating a copy of an event, we do not want to store/remove this in/from the persistence backend.
            When the original event is requeued it will pass through all modules again, thus creating the
            same copies etc.
            A removeFromPersistenceBackendOnGarbageCollect would fail anyways, since the new key is unknown to the
            backend. Still we can skip removing for speedups.
            """
            new_dict = DictUtils.KeyDotNotationDict()
            new_dict.__init__ = DictUtils.KeyDotNotationDict.___init___
            new_dict.__del__ = DictUtils.KeyDotNotationDict.__del__
            new_dict.update(copy.deepcopy(super(DictUtils.KeyDotNotationDict, self)))
            if "event_id" in new_dict.get("lumbermill", {}):
                new_dict['lumbermill']['event_id'] = "%032x" % random.getrandbits(128)
            return new_dict
        DictUtils.KeyDotNotationDict.copy = returnSimpleKeyDotDictOnCopy

    def getTimedGarbageCollectFunc(self):
        @setInterval(self.flush_interval, call_on_init=True)
        def timedGarbageCollect():
            if not self.requeue_events_done:
                self.requeueEvents()
                self.requeue_events_done = True
            gc.collect()
        return timedGarbageCollect

    def requeueEvents(self):
        input_modules = {}
        for module_name, module_info in self.lumbermill.modules.items():
            instance = module_info['instances'][0]
            if instance.module_type == "input":
                input_modules[instance.__class__.__name__] = instance
        for key in self.persistence_backend.iterKeys():
            if not key.startswith("%s" % self.key_prefix):
                continue
            self.logger.warning("Found %s unfinished events. Requeing..." % (len(keys)))
            requeue_counter = 0
            for key in keys:
                event = self.persistence_backend.pop(key)
                if not event:
                    continue
                if "source_module" not in event.get("lumbermill", {}):
                    self.logger.warning("Could not requeue event. Source module info not found in event data.")
                    continue
                source_module = event["lumbermill"]["source_module"]
                if source_module not in input_modules:
                    self.logger.error("Could not requeue event. Module %s not found." % (source_module))
                    continue
                requeue_counter += 1
                input_modules[source_module].sendEvent(DictUtils.KeyDotNotationDict(event))
            self.logger.warning("Done. Requeued %s of %s events." % (requeue_counter, len(keys)))
            self.logger.warning("Note: If more than one gp instance is running, requeued events count may differ from total events.")
            event = None

    def start(self):
        self.timedFuncHandle = TimedFunctionManager.startTimedFunction(self.getTimedGarbageCollectFunc())


# -*- coding: utf-8 -*-
from lumbermill.BaseThreadedModule import BaseThreadedModule
from lumbermill.utils.Decorators import ModuleDocstringParser


@ModuleDocstringParser
class DropEvent(BaseThreadedModule):
    """
    Drop all events received by this module.

    This module is intended to be used with an activated filter.

    Configuration template:

    - DropEvent
    """

    module_type = "modifier"
    """Set module type"""

    def handleEvent(self, event):
        yield None


# -*- coding: utf-8 -*-
import sys

import lumbermill.utils.DictUtils as DictUtils
from lumbermill.BaseThreadedModule import BaseThreadedModule
from lumbermill.utils.Decorators import ModuleDocstringParser, setInterval
from lumbermill.utils.misc import TimedFunctionManager


@ModuleDocstringParser
class Facet(BaseThreadedModule):
    """
    Collect different values of one field over a defined period of time and pass all
    encountered variations on as new event after period is expired.

    The "add_event_fields" configuration will copy the configured event fields into the "other_event_fields" list.

    The event emitted by this module will be of type: "facet" and will have "facet_field",
    "facet_count", "facets" and "other_event_fields" fields set.

    This module supports the storage of the facet info in an redis db. If redis_store is set,
    it will first try to retrieve the facet info from redis via the key setting.

    Configuration template:

    - Facet:
       source_field:                    # <type:string; is: required>
       group_by:                        # <type:string; is: required>
       add_event_fields:                # <default: []; type: list; is: optional>
       interval:                        # <default: 5; type: float||integer; is: optional>
       redis_store:                     # <default: None; type: None||string; is: optional>
       redis_ttl:                       # <default: 60; type: integer; is: optional>
       receivers:
        - NextModule
    """
    facet_data = {}
    """Holds the facet data for all instances"""

    redis_keys = []
    """Holds just the redis keys to all facet data"""

    module_type = "misc"
    """Set module type"""

    def configure(self, configuration):
        # Call parent configure method
        BaseThreadedModule.configure(self, configuration)
        self.source_field = self.getConfigurationValue('source_field')
        self.add_event_fields = self.getConfigurationValue('add_event_fields')
        # Get redis client module.
        if self.getConfigurationValue('redis_store'):
            mod_info = self.lumbermill.getModuleInfoById(self.getConfigurationValue('redis_store'))
            self.redis_store = mod_info['instances'][0]
        else:
            self.redis_store = None

    def _getFacetInfoRedis(self, key):
        facet_info = self.redis_store.get(key)
        if not facet_info:
            facet_info = {'other_event_fields': {}, 'facets': []}
        return facet_info

    def _getFacetInfoInternal(self, key):
        try:
            facet_info = Facet.facet_data[key]
        except KeyError:
            facet_info = {'other_event_fields': {}, 'facets': []}
        return facet_info

    def getFacetInfo(self, key):
        if self.redis_store:
            return self._getFacetInfoRedis(key)
        return self._getFacetInfoInternal(key)

    def _setFacetInfoRedis(self, key, facet_info):
        try:
            self.redis_store.set(key, facet_info, self.getConfigurationValue('redis_ttl'))
            if key not in Facet.redis_keys:
                Facet.redis_keys.append(key)
        except:
            etype, evalue, etb = sys.exc_info()
            self.logger.warning("Could not store facet data in redis. Exception: %s, Error: %s." % (etype, evalue))
            pass

    def _setFacetInfoInternal(self, key, facet_info):
        Facet.facet_data[key] = facet_info

    def setFacetInfo(self, key, facet_info):
        if self.redis_store:
            self._setFacetInfoRedis(key, facet_info)
            return
        self._setFacetInfoInternal(key, facet_info)

    def sendFacetEventToReceivers(self, facet_data):
        event = DictUtils.getDefaultEventDict({'facet_field': self.source_field,
                                          'facet_count': len(facet_data['facets']),
                                          'facets': facet_data['facets']},
                                          caller_class_name=self.__class__.__name__,
                                          event_type='facet')
        if facet_data['other_event_fields']:
            event['other_event_fields'] = facet_data['other_event_fields']
        self.sendEvent(event)

    def getEvaluateFunc(self):
        @setInterval(self.getConfigurationValue('interval'))
        def evaluateFacets():
            if self.redis_store:
                for key in Facet.redis_keys:
                    self.sendFacetEventToReceivers(self._getFacetInfoRedis(key))
                    # Clear redis items
                    self.redis_store.client.delete(key)
                Facet.redis_keys = []
                return
            # Just internal facet data.
            for key, facet_data in Facet.facet_data.items():
                self.sendFacetEventToReceivers(facet_data)
            Facet.facet_data = {}
        return evaluateFacets

    def initAfterFork(self):
        self.evaluate_facet_data_func = self.getEvaluateFunc()
        self.timed_func_handler = TimedFunctionManager.startTimedFunction(self.evaluate_facet_data_func)
        BaseThreadedModule.initAfterFork(self)

    def handleEvent(self, event):
        """
        Process the event.

        @param event: dictionary
        @return data: dictionary
        """
        # Get key and value for facet from event.
        try:
            facet_value = event[self.getConfigurationValue('source_field', event)]
        except KeyError:
            yield event
            return
        key = self.getConfigurationValue('group_by', event)
        if not key and self.redis_store:
            self.logger.warning("Group_by value %s could not be generated. Event ignored." % (self.getConfigurationValue('group_by')))
            yield event
            return
        key = "FacetValues:%s" % key
        redis_lock = False
        try:
            # Acquire redis lock as well if configured to use redis store.
            if self.redis_store:
                redis_lock = self.redis_store.getLock("FacetLocks:%s" % key, timeout=1)
                if not redis_lock:
                    yield event
                    return
                redis_lock.acquire()
            facet_info = self.getFacetInfo(key)
            if facet_value not in facet_info['facets']:
                keep_fields = {}
                for field_name in self.add_event_fields:
                    try:
                        keep_fields[field_name] = event[field_name]
                    except KeyError:
                        pass
                if keep_fields:
                    facet_info['other_event_fields'][facet_value] = keep_fields
                facet_info['facets'].append(facet_value)
                self.setFacetInfo(key, facet_info)
        except:
            # Pass on all exceptions
            raise
        finally:
            # Make sure redis lock is released if we have one.
            if redis_lock:
                redis_lock.release()
        yield event

    def shutDown(self):
        # Push any remaining facet data.
        self.evaluate_facet_data_func(self)
        # Call parent configure method.
        BaseThreadedModule.shutDown(self)


# -*- coding: utf-8 -*-
import sys

import redis

from lumbermill.BaseThreadedModule import BaseThreadedModule
from lumbermill.utils.Decorators import ModuleDocstringParser
from lumbermill.utils.DynamicValues import mapDynamicValue


@ModuleDocstringParser
class RedisChannelSink(BaseThreadedModule):
    """
    Publish incoming events to redis channel.

    channel: Name of redis channel to send data to.
    server: Redis server to connect to.
    port: Port redis server is listening on.
    db: Redis db.
    password: Redis password.
    format: Which event fields to send on, e.g. '$(@timestamp) - $(url) - $(country_code)'. If not set, the whole event dict is send.

    Configuration template:

    - RedisChannelSink:
       channel:                         # <type: string; is: required>
       server:                          # <default: 'localhost'; type: string; is: optional>
       port:                            # <default: 6379; type: integer; is: optional>
       db:                              # <default: 0; type: integer; is: optional>
       password:                        # <default: None; type: None||string; is: optional>
       format:                          # <default: None; type: None||string; is: optional>
       store_interval_in_secs:          # <default: 5; type: integer; is: optional>
       batch_size:                      # <default: 500; type: integer; is: optional>
       backlog_size:                    # <default: 5000; type: integer; is: optional>
    """

    module_type = "output"
    """Set module type"""
    can_run_forked = True

    def configure(self, configuration):
         # Call parent configure method
        BaseThreadedModule.configure(self, configuration)
        self.format = self.getConfigurationValue('format')
        try:
            self.client = redis.Redis(host=self.getConfigurationValue('server'),
                                      port=self.getConfigurationValue('port'),
                                      password=self.getConfigurationValue('password'),
                                      db=self.getConfigurationValue('db'))
        except:
            etype, evalue, etb = sys.exc_info()
            self.logger.error("Could not connect to redis store at %s. Exception: %s, Error: %s." % (self.getConfigurationValue('server'),etype, evalue))

    def initAfterFork(self):
        #self.buffer = Utils.Buffer(self.getConfigurationValue('batch_size'), self.storeData, self.getConfigurationValue('store_interval_in_secs'), maxsize=self.getConfigurationValue('backlog_size'))
        BaseThreadedModule.initAfterFork(self)

    def handleEvent(self, event):
        if self.format:
            publish_data = mapDynamicValue(self.format, event)
        else:
            publish_data = event
        try:
            self.client.publish(self.getConfigurationValue('channel', event), publish_data)
        except:
            etype, evalue, etb = sys.exc_info()
            self.logger.error("Could not publish event to redis channel %s at %s. Exception: %s, Error: %s." % (self.getConfigurationValue('channel', event), self.getConfigurationValue('server'), etype, evalue))
        yield None

    def __handleEvent(self, event):
        if self.format:
            publish_data = mapDynamicValue(self.format, event)
        else:
            publish_data = event
        self.buffer.append(publish_data)
        yield None


# -*- coding: utf-8 -*-
import socket
import sys

import zmq
from zmq.eventloop import ioloop, zmqstream

ioloop.install()

import lumbermill.utils.DictUtils as DictUtils
from lumbermill.BaseModule import BaseModule
from lumbermill.utils.Decorators import ModuleDocstringParser


@ModuleDocstringParser
class ZmqTornado(BaseModule):
    """
    Read events from a zeromq.

    mode: Whether to run a server or client.
    address: Address to connect to. Pattern: hostname:port. If mode is server, this sets the addresses to listen on.
    pattern: One of 'pull', 'sub'.
    hwm: Highwatermark for sending/receiving socket.
    separator: When using the sub pattern, messages can have a topic. Set separator to split message from topic.

    Configuration template:

    - ZmqTornado:
       mode:                            # <default: 'server'; type: string; values: ['server', 'client']; is: optional>
       address:                         # <default: '*:5570'; type: string; is: optional>
       pattern:                         # <default: 'pull'; type: string; values: ['pull', 'sub']; is: optional>
       topic:                           # <default: ''; type: string; is: optional>
       separator:                       # <default: None; type: None||string; is: optional>
       hwm:                             # <default: None; type: None||integer; is: optional>
       receivers:
        - NextModule
    """

    module_type = "input"
    """Set module type"""
    can_run_forked = False

    zmq_pattern_mapping = {'push': zmq.PUSH,
                           'pull': zmq.PULL,
                           'pub': zmq.PUB,
                           'sub': zmq.SUB}

    def configure(self, configuration):
         # Call parent configure method
        BaseModule.configure(self, configuration)
        self.client = None
        self.topic = self.getConfigurationValue('topic')
        self.separator = self.getConfigurationValue('separator')
        self.context = zmq.Context()
        self.socket = self.context.socket(self.zmq_pattern_mapping[self.getConfigurationValue('pattern')])
        if self.getConfigurationValue('hwm'):
            self.setReceiveHighWaterMark(self.getConfigurationValue('hwm'))
        server_addr, server_port = self.getServerAddress(self.getConfigurationValue('address'))
        if self.getConfigurationValue('mode') == 'server':
            self.bind(server_addr, server_port)
        else:
            self.connect(server_addr, server_port)
        self.socket = zmqstream.ZMQStream(self.socket)
        self.socket.on_recv(self.onReceive)

    def getServerAddress(self, server_name):
        # ZMQ does not like hostnames too much. Try to get the ip address for server name.
        server_name, server_port = server_name.split(":")
        try:
            server_addr = socket.gethostbyname(server_name)
        except socket.gaierror:
            server_addr = server_name
        return (server_addr, server_port)

    def setReceiveHighWaterMark(self, hwm):
        try:
            self.socket.setsockopt(zmq.RCVHWM, hwm)
        except:
            self.socket.setsockopt(zmq.HWM, hwm)

    def bind(self, server_addr, server_port):
        try:
            self.socket.bind('tcp://%s:%s' % (server_addr, server_port))
        except:
            etype, evalue, etb = sys.exc_info()
            self.logger.error("Could not bind to %s:%s. Exception: %s, Error: %s." % (server_addr, server_port, etype, evalue))
            self.lumbermill.shutDown()

    def connect(self, server_addr, server_port):
        try:
            self.socket.connect('tcp://%s:%s' % (server_addr, server_port))
        except:
            etype, evalue, etb = sys.exc_info()
            self.logger.error("Could not connect to zeromq at %s:%s. Exception: %s, Error: %s." % (server_addr, server_port, etype, evalue))
            self.lumbermill.shutDown()

    def onReceive(self, data):
        data = data[0]
        if self.separator:
            topic, data = data.split(self.separator)
        event = DictUtils.getDefaultEventDict({"data": data}, caller_class_name="ZmqTornado")
        self.sendEvent(event)

    def shutDown(self):
        try:
            self.socket.close()
            self.context.term()
        except AttributeError:
            pass
        # Call parent shutDown method.
        BaseModule.shutDown(self)


# -*- coding: utf-8 -*-
import csv
import sys
from cStringIO import StringIO

from lumbermill.BaseThreadedModule import BaseThreadedModule
from lumbermill.utils.Decorators import ModuleDocstringParser


@ModuleDocstringParser
class CsvParser(BaseThreadedModule):
    """
    Parse a string as csv data.

    It will parse the csv and create or replace fields in the internal data dictionary with
    the corresponding csv fields.

    source_field: Field that contains the csv data.
    escapechar: Char used to escape special characters.
    skipinitialspace: When True, whitespace immediately following the delimiter is ignored. The default is False.
    quotechar: A one-character string used to quote fields containing special characters, such as the delimiter or quotechar, or which contain new-line characters.
    delimiter: A one-character string used to separate fields.
    fieldnames: Fieldnames to be used for the extracted csv data.

    Configuration template:

    - CsvParser:
       source_field:                    # <default: 'data'; type: string; is: optional>
       escapechar:                      # <default: '\'; type: string; is: optional>
       skipinitialspace:                # <default: False; type: boolean; is: optional>
       quotechar:                       # <default: '"'; type: string; is: optional>
       delimiter:                       # <default: '|'; type: string; is: optional>
       fieldnames:                      # <type: list; is: required>
       receivers:
        - NextModule
    """

    module_type = "parser"
    """Set module type"""

    def handleEvent(self, event):
        try:
            csv_dict = csv.reader(StringIO(event[self.getConfigurationValue('source_field', event)]),
                                  escapechar=self.getConfigurationValue('escapechar', event),
                                  skipinitialspace=self.getConfigurationValue('skipinitialspace', event),
                                  quotechar=self.getConfigurationValue('quotechar', event),
                                  delimiter=self.getConfigurationValue('delimiter', event))
        except:
            etype, evalue, etb = sys.exc_info()
            self.logger.warning("Could not parse csv data %s. Exception: %s, Error: %s." % (event, etype, evalue))
            yield event
            return
        field_names = self.getConfigurationValue('fieldnames', event)
        for values in csv_dict:
            for index, value in enumerate(values):
                try:
                    event[field_names[index]] = value
                except KeyError:
                    pass
                except IndexError:
                    pass
        yield event


# -*- coding: utf-8 -*-
import copy
import os
import random
from string import Formatter

from lumbermill.constants import MY_HOSTNAME

class KeyDotNotationDict(dict):
    """
    A dictionary that allows to access values via dot separated keys, e.g.:
    >>> my_dict = {"key1": {"key2": "value"}}
    >>> my_dict["key1.key2"]
    "value"
    """

    def __getitem__(self, key, dict_or_list=None):
        dict_or_list = dict_or_list if dict_or_list else super(KeyDotNotationDict, self)
        if "." not in key:
            if isinstance(dict_or_list, list):
                key = int(key)
            return dict_or_list.__getitem__(key)
        current_key, remaining_keys = key.split('.', 1)
        try:
            dict_or_list = dict_or_list.__getitem__(current_key)
        except TypeError:
            dict_or_list = dict_or_list.__getitem__(int(current_key))
        return self.__getitem__(remaining_keys, dict_or_list)

    def __setitem__(self, key, value, dict_or_list=None):
        dict_or_list = dict_or_list if dict_or_list else super(KeyDotNotationDict, self)
        if "." not in key:
            if isinstance(dict_or_list, list):
                key = int(key)
            return dict_or_list.__setitem__(key, value)
        current_key, remaining_keys = key.split('.', 1)
        dict_or_list = dict_or_list.__getitem__(current_key)
        return self.__setitem__(remaining_keys, value, dict_or_list)

    def __delitem__(self, key, dict_or_list=None):
        dict_or_list = dict_or_list if dict_or_list else super(KeyDotNotationDict, self)
        if "." not in key:
            if isinstance(dict_or_list, list):
                key = int(key)
            return dict_or_list.__delitem__(key)
        current_key, remaining_keys = key.split('.', 1)
        dict_or_list = dict_or_list.__getitem__(current_key)
        return self.__delitem__(remaining_keys, dict_or_list)

    def __contains__(self, key, dict_or_list=None):
        dict_or_list = dict_or_list if dict_or_list else super(KeyDotNotationDict, self)
        if "." not in key:
            return dict_or_list.__contains__(key)
        current_key, remaining_keys = key.split('.', 1)
        try:
            dict_or_list = dict_or_list.__getitem__(current_key)
            return self.__contains__(remaining_keys, dict_or_list)
        except KeyError:
            return False

    #def __del__(self, key):
    #    pass

    def __missing__(self, key):
        raise KeyError(key)

    def copy(self):
        new_dict = KeyDotNotationDict(copy.deepcopy(super(KeyDotNotationDict, self)))
        if "event_id" in new_dict.get("lumbermill", {}):
            new_dict['lumbermill']['event_id'] = "%032x%s" % (random.getrandbits(128), os.getpid())
        return new_dict

    def get(self, key, *args):
        try:
            return self.__getitem__(key)
        except (KeyError, IndexError) as e:
            # Try to return a default value for not existing dict keys.
            try:
                return args[0]
            except KeyError:
                raise e

    def __get(self, key, default, dict_or_list=None):
        dict_or_list = dict_or_list if dict_or_list else super(KeyDotNotationDict, self)
        if "." not in key:
            if not isinstance(dict_or_list, list):
                return dict_or_list.get(key, default)
            else:
                try:
                    return dict_or_list[int(key)]
                except KeyError:
                    return default
        current_key, remaining_keys = key.split('.', 1)
        try:
            dict_or_list = dict_or_list.__getitem__(current_key)
            return self.get(remaining_keys, default, dict_or_list)
        except KeyError:
            return default

    def pop(self, key, default=None, dict_or_list=None):
        dict_or_list = dict_or_list if dict_or_list else super(KeyDotNotationDict, self)
        if "." not in key:
            if not isinstance(dict_or_list, list):
                return dict_or_list.pop(key, default)
            else:
                try:
                    return dict_or_list[int(key)]
                except KeyError:
                    return default
        current_key, remaining_keys = key.split('.', 1)
        try:
            dict_or_list = dict_or_list.__getitem__(current_key)
            return self.pop(remaining_keys, default, dict_or_list)
        except KeyError:
            return default

class DotDictFormatter(Formatter):
    try:  # deal with Py 2 & 3 difference
        NUMERICS = (int, long)
    except NameError:
        NUMERICS = int

    def get_field(self, field_name, args, kwargs):
        first, rest = field_name._formatter_field_name_split()
        obj = self.get_value(first, args, kwargs)
        # loop through the rest of the field_name, doing
        #  getattr or getitem as needed
        for is_attr, i in rest:
            obj = obj[i]
        return obj, first

def mergeNestedDicts(a, b, path=None):
    "Merges a with b. If b provides same key as a, b takes precendence."
    if path is None: path = []
    for key in b:
        if key in a:
            if isinstance(a[key], dict) and isinstance(b[key], dict):
                mergeNestedDicts(a[key], b[key], path + [str(key)])
            elif a[key] == b[key]:
                continue # same leaf value
            else:
                a[key] = b[key]
        else:
            a[key] = b[key]
    return a

def getDefaultEventDict(dict={}, caller_class_name='', received_from=False, event_type="Unknown"):
    default_dict = { "data": "",
                     "lumbermill": {
                         'pid': os.getpid(),
                         'event_type': event_type,
                         'event_id': "%032x%s" % (random.getrandbits(128), os.getpid()),
                         'source_module': caller_class_name,
                         'received_from': received_from,
                         'received_by': MY_HOSTNAME
                     }
                }
    default_dict.update(dict)
    default_dict = KeyDotNotationDict(default_dict)
    return default_dict


# -*- coding: utf-8 -*-
import Queue
import SocketServer
import logging
import socket
import sys
import threading

import lumbermill.utils.DictUtils as DictUtils
from lumbermill.BaseModule import BaseModule
from lumbermill.utils.Decorators import ModuleDocstringParser


class ThreadPoolMixIn(SocketServer.ThreadingMixIn):
    """
    Use a thread pool instead of a new thread on every request.

    Using a threadpool prevents the spawning of a new thread for each incoming
    request. This should increase performance a bit.

    See: http://code.activestate.com/recipes/574454/
    """
    numThreads = 15
    allow_reuse_address = True  # seems to fix socket.error on server restart
    alive = True

    def serve_forever(self):
        """
        Handle one request at a time until doomsday.
        """
        # Set up the threadpool.
        self.requests = Queue.Queue(self.numThreads)

        for x in range(self.numThreads):
            t = threading.Thread(target=self.process_request_thread)
            t.setDaemon(1)
            t.start()

        # server main loop
        while self.alive:
            self.handle_request()
        self.server_close()


    def process_request_thread(self):
        """
        obtain request from queue instead of directly from server socket
        """
        while True:
            SocketServer.ThreadingMixIn.process_request_thread(self, *self.requests.get())


    def handle_request(self):
        """
        simply collect requests and put them on the queue for the workers.
        """
        try:
            request, client_address = self.get_request()
        except:
            etype, evalue, etb = sys.exc_info()
            print "Exception: %s, Error: %s." % (etype, evalue)
            return
        #if self.verify_request(request, client_address):
        self.requests.put((request, client_address))


class ThreadedUdpRequestHandler(SocketServer.BaseRequestHandler):

    def __init__(self, udp_server_instance, *args, **keys):
        self.udp_server_instance = udp_server_instance
        self.logger = logging.getLogger(self.__class__.__name__)
        SocketServer.BaseRequestHandler.__init__(self, *args, **keys)

    def handle(self):
        try:
            data = self.request[0].strip()
            if data == "":
                return
            host = self.client_address[0]
            port = self.client_address[1]
            event = DictUtils.getDefaultEventDict({"data": data}, received_from="%s:%s" % (host, port),caller_class_name='UdpServer')
            self.udp_server_instance.sendEvent(event)
        except socket.error, e:
           self.logger.warning("Error occurred while reading from socket. Error: %s" % (e))
        except socket.timeout, e:
            self.logger.warning("Timeout occurred while reading from socket. Error: %s" % (e))

class ThreadedUdpServer(ThreadPoolMixIn, SocketServer.UDPServer):

    allow_reuse_address = True

    def __init__(self, server_address, RequestHandlerClass, bind_and_activate=True, timeout=None):
        SocketServer.UDPServer.__init__(self, server_address, RequestHandlerClass)
        self.socket.settimeout(timeout)
        self.timeout = timeout

class UdpRequestHandlerFactory:
    def produce(self, tcp_server_instance):
        def createHandler(*args, **keys):
            return ThreadedUdpRequestHandler(tcp_server_instance, *args, **keys)
        return createHandler

@ModuleDocstringParser
class UdpServer(BaseModule):
    """
    Reads data from udp socket and sends it to its output queues.

    interface:  Ipaddress to listen on.
    port:       Port to listen on.
    timeout:    Sockettimeout in seconds.

    Configuration template:

    - UdpServer:
       interface:                       # <default: '0.0.0.0'; type: string; is: optional>
       port:                            # <default: 5151; type: integer; is: optional>
       timeout:                         # <default: None; type: None||integer; is: optional>
       receivers:
        - NextModule
    """

    module_type = "input"
    """Set module type"""

    can_run_forked = True

    def configure(self, configuration):
        # Call parent configure method
        BaseModule.configure(self, configuration)
        self.server = False
        handler_factory = UdpRequestHandlerFactory()
        try:
            self.server = ThreadedUdpServer((self.getConfigurationValue("interface"),
                                             self.getConfigurationValue("port")),
                                             handler_factory.produce(self),
                                             timeout=self.getConfigurationValue("timeout"))
        except:
            etype, evalue, etb = sys.exc_info()
            self.logger.error("Could not listen on %s:%s. Exception: %s, Error: %s" % (self.getConfigurationValue("interface"),
                                                                                       self.getConfigurationValue("port"), etype, evalue))
            self.lumbermill.shutDown()
            return

    def getStartMessage(self):
        """
        Return the module name.
        """
        return "listening on %s:%s" % (self.getConfigurationValue("interface"), self.getConfigurationValue("port"))

    def start(self):
        if not self.receivers:
            self.logger.error("Shutting down module %s since no receivers are set." % (self.__class__.__name__))
            return
        # Start a thread with the server -- that thread will then start one
        # more threads for each request.
        server_thread = threading.Thread(target=self.server.serve_forever)
        # Exit the server thread when the main thread terminates.
        server_thread.daemon = True
        server_thread.start()

    def shutDown(self):
        try:
            self.server.alive = False
        except AttributeError:
            pass



# -*- coding: utf-8 -*-
import types

from ua_parser import user_agent_parser

from lumbermill.BaseThreadedModule import BaseThreadedModule
from lumbermill.utils.Buffers import MemoryCache
from lumbermill.utils.Decorators import ModuleDocstringParser


@ModuleDocstringParser
class UserAgentParser(BaseThreadedModule):
    r"""
    Parse http user agent string

    A string like:

        "Mozilla/5.0 (Linux; U; Android 2.3.5; en-in; HTC_DesireS_S510e Build/GRJ90) AppleWebKit/533.1 (KHTML, like Gecko) Version/4.0 Mobile Safari/533.1"

    will produce this dictionary:

    'user_agent_info': {   'device': {   'family': u'HTC DesireS'},
                           'os': {   'family': 'Android',
                                     'major': '2',
                                     'minor': '3',
                                     'patch': '5',
                                     'patch_minor': None},
                           'user_agent': {   'family': 'Android',
                                             'major': '2',
                                             'minor': '3',
                                             'patch': '5'}}}

    source_fields:  Input field to parse.
    target_field: field to update with parsed info fields.

    Configuration template:

    - UserAgentParser:
       source_fields:                   # <type: string||list; is: required>
       target_field:                    # <default: 'user_agent_info'; type:string; is: optional>
       receivers:
        - NextModule
    """

    module_type = "parser"
    """Set module type"""

    def configure(self, configuration):
        # Call parent configure method
        BaseThreadedModule.configure(self, configuration)
        self.source_fields = self.getConfigurationValue('source_fields')
        # Allow single string as well.
        if isinstance(self.source_fields, types.StringTypes):
            self.source_fields = [self.source_fields]
        self.target_field = self.getConfigurationValue('target_field')
        self.in_mem_cache = MemoryCache(size=1000)

    def handleEvent(self, event):
        for source_field in self.source_fields:
            if source_field not in event:
                continue
            # Try to get it from cache.
            try:
                ua_info = self.in_mem_cache.get(event[source_field])
            except KeyError:
                # Drop the 'string' field to avoid duplicate data.
                ua_info = user_agent_parser.Parse(event[source_field])
                if 'string' in ua_info:
                    ua_info.pop('string')
                self.in_mem_cache.set(event[source_field], ua_info)
            event[self.target_field] = ua_info
        yield event



# -*- coding: utf-8 -*-
import os.path
import socket
import sys
import time

import tornado.autoreload
import tornado.httpserver
import tornado.ioloop
import tornado.web

from lumbermill.BaseThreadedModule import BaseThreadedModule
from lumbermill.utils.Decorators import ModuleDocstringParser


@ModuleDocstringParser
class WebserverTornado(BaseThreadedModule):
    """
    A tornado based web server.

    Configuration template:

    - WebserverTornado:
       port:                            # <default: 5100; type: integer; is: optional>
       tls:                             # <default: False; type: boolean; is: optional>
       key:                             # <default: False; type: boolean||string; is: required if tls is True else optional>
       cert:                            # <default: False; type: boolean||string; is: required if tls is True else optional>
       document_root:                   # <default: '../assets/webserver_docroot'; type: string; is: optional>
       application_settings:            # <default: None; type: None||dict; is: optional>
    """
    module_type = "stand_alone"
    """Set module type"""

    can_run_forked = False

    def configure(self, configuration):
        # Call parent configure method
        BaseThreadedModule.configure(self, configuration)
        self.server = False
        self.settings = self.getSettings()
        self.application = tornado.web.Application([], **self.settings)

    def addHandlers(self, host_handlers=[], host_pattern='.*$'):
        self.application.add_handlers(host_pattern, host_handlers)

    def getSettings(self):
        base_path = self.getConfigurationValue('document_root')
        if base_path == 'docroot':
            base_path = "%s/docroot" % os.path.dirname(__file__)
        settings = {'template_path' : "%s/templates" % base_path,
                    'static_path': "%s/static" % base_path,
                    'ui_modules': [],
                    'debug': False,
                    'TornadoWebserver': self}
        if self.getConfigurationValue('application_settings'):
            settings.update(self.getConfigurationValue('application_settings'))
        return settings

    def addUiModules(self, modules):
        if not isinstance(modules, list):
            modules = [modules]
        for module in modules:
            if module in self.application.settings['ui_modules']:
                continue
            self.application.settings['ui_modules'].append(module)
            self.application._load_ui_modules(module)

    def getUiModule(self, module_name):
        try:
            return self.application.ui_modules[module_name]
        except KeyError:
            return False

    def start(self):
        ssl_options = None
        if self.getConfigurationValue("tls"):
            ssl_options = { 'certfile': self.getConfigurationValue("cert"),
                            'keyfile': self.getConfigurationValue("key")}
        try:
            self.server = tornado.httpserver.HTTPServer(self.application, ssl_options=ssl_options)
            self.server.listen(self.getConfigurationValue('port'))
            for fd, server_socket in self.server._sockets.items():
                server_socket.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
        except:
            etype, evalue, etb = sys.exc_info()
            self.logger.error("Could not start webserver on %s. Exception: %s, Error: %s." % (self.getConfigurationValue('port'), etype, evalue))
            return
        tornado.autoreload.add_reload_hook(self.shutDown)
        return
        ioloop = tornado.ioloop.IOLoop.instance()
        ioloop.make_current()
        try:
            ioloop.start()
        except ValueError:
            # Ignore errors like "ValueError: I/O operation on closed kqueue fd". These might be thrown during a reload.
            pass

    def shutDown(self):
        if self.server:
            self.server.stop()
            # Give os time to free the socket. Otherwise a reload will fail with 'address already in use'
            time.sleep(.2)
        # Call parent shutDown method.
        BaseThreadedModule.shutDown(self)



# -*- coding: utf-8 -*-
import time

from lumbermill.BaseThreadedModule import BaseThreadedModule
from lumbermill.utils.Decorators import ModuleDocstringParser


@ModuleDocstringParser
class LoggerSink(BaseThreadedModule):
    """
    Send data to lumbermill logger.

    formats: Format of messages to send to logger, e.g.:
             ['############# Statistics #############',
              'Received events in $(interval)s: $(total_count)',
              'EventType: httpd_access_log - Hits: $(field_counts.httpd_access_log)',
              'EventType: Unknown - Hits: $(field_counts.Unknown)']

    Configuration template:

    - LoggerSink:
       formats:                         # <type: list; is: required>
    """

    module_type = "output"
    """Set module type"""

    def configure(self, configuration):
        # Call parent configure method
        BaseThreadedModule.configure(self, configuration)
        self.formats = self.getConfigurationValue('formats')
        self.printing = False

    def handleEvent(self, event):
        while self.printing:
            time.sleep(.0001)
        self.printing = True
        for format in self.formats:
            output = self.mapDynamicValue(format, event)
            if not output:
                continue
            self.logger.info("%s" % (output))
        self.printing = False
        yield None


# -*- coding: utf-8 -*-
import re
import sys
import types
from json import JSONDecoder

from lumbermill.constants import IS_PYPY
from lumbermill.BaseThreadedModule import BaseThreadedModule
from lumbermill.utils.Decorators import ModuleDocstringParser

# For pypy the default json module is the fastest.
if IS_PYPY:
    import json
else:
    json = False
    for module_name in ['ujson', 'yajl', 'simplejson', 'json']:
        try:
            json = __import__(module_name)
            break
        except ImportError:
            pass
    if not json:
        raise ImportError

#shameless copy paste from json/decoder.py
FLAGS = re.VERBOSE | re.MULTILINE | re.DOTALL
WHITESPACE = re.compile(r'[ \t\n\r]*', FLAGS)

class ConcatJSONDecoder(JSONDecoder): # json.JSONDecoder
    def decode(self, s, _w=WHITESPACE.match):
        s_len = len(s)
        objs = []
        end = 0
        while end != s_len:
            obj, end = self.raw_decode(s, idx=_w(s, end).end())
            end = _w(s, end).end()
            objs.append(obj)
        return objs

@ModuleDocstringParser
class JsonParser(BaseThreadedModule):
    """
    Json codec.

    Decode:
    It will parse the json data in source fields and create or replace fields in the internal data dictionary with
    the corresponding json fields.

    Encode:
    It will build a new list of source fields and create json of this list.

    At the moment only flat json files can be processed correctly.

    action:         Either encode or decode data.
    source_fields:  Input fields for de/encode.
                    If encoding, you can set this field to 'all' to encode the complete event dict.
    target_field:   Target field for de/encode result.
                    If decoding and target is not set, the event dict itself will be updated with decoded fields.
    keep_original:  Switch to keep or drop the original fields used in de/encoding from the event dict.

    Configuration template:

    - JsonParser:
       action:                          # <default: 'decode'; type: string; values: ['decode','encode']; is: optional>
       mode:                            # <default: 'line'; type: string; values: ['line','stream']; is: optional>
       source_fields:                   # <default: 'data'; type: string||list; is: optional>
       target_field:                    # <default: None; type: None||string; is: optional>
       keep_original:                   # <default: False; type: boolean; is: optional>
       receivers:
        - NextModule
    """

    module_type = "parser"
    """Set module type"""

    def configure(self, configuration):
        # Call parent configure method
        BaseThreadedModule.configure(self, configuration)
        self.source_fields = self.getConfigurationValue('source_fields')
        # Allow single string as well.
        if isinstance(self.source_fields, types.StringTypes):
            self.source_fields = [self.source_fields]
        self.target_field = self.getConfigurationValue('target_field')
        self.drop_original = not self.getConfigurationValue('keep_original')
        if self.getConfigurationValue('action') == 'decode':
            if self.getConfigurationValue('mode') == 'line':
                self.handleEvent = self.decodeEventLine
            else:
                self.handleEvent = self.decodeEventStream
        else:
            self.handleEvent = self.encodeEvent

    def decodeEventLine(self, event):
        for source_field in self.source_fields:
            if source_field not in event:
                continue
            json_string = str(event[source_field]).strip("'<>() ").replace('\'', '\"')
            decoded_datasets = ""
            try:
                decoded_datasets = json.loads(json_string)
            except:
                # Maybe we got a stream of multiple json messages. Try to parse them.
                try:
                    decoded_datasets = json.loads(json_string, cls=ConcatJSONDecoder)
                except:
                    etype, evalue, etb = sys.exc_info()
                    self.logger.warning("Could not json decode event data: %s. Exception: %s, Error: %s." % (event, etype, evalue))
                    continue
            if not isinstance(decoded_datasets, list):
                decoded_datasets = [decoded_datasets]
            copy_event = False
            for decoded_data in decoded_datasets:
                if copy_event:
                    event = event.copy()
                copy_event = True
                if self.drop_original:
                    event.pop(source_field, None)
                if self.target_field:
                    event.update({self.target_field: decoded_data})
                else:
                    event.update(decoded_data)
                yield event

    def decodeEventStream(self, event):
        for source_field in self.source_fields:
            if source_field not in event:
                continue
            json_string = str(event[source_field]).strip("'<>() ").replace('\'', '\"')
            decoded_datasets = ""
            try:
                decoded_datasets = json.loads(json_string)
            except:
                # Maybe we got a stream of multiple json messages. Try to parse them.
                try:
                    decoded_datasets = json.loads(json_string, cls=ConcatJSONDecoder)
                except:
                    etype, evalue, etb = sys.exc_info()
                    self.logger.warning("Could not json decode event data: %s. Exception: %s, Error: %s." % (event, etype, evalue))
                    continue
            if not isinstance(decoded_datasets, list):
                decoded_datasets = [decoded_datasets]
            copy_event = False
            for decoded_data in decoded_datasets:
                if copy_event:
                    event = event.copy()
                copy_event = True
                if self.drop_original:
                    event.pop(source_field, None)
                if self.target_field:
                    event.update({self.target_field: decoded_data})
                else:
                    event.update(decoded_data)
                yield event

    def encodeEvent(self, event):
        if 'all' in self.source_fields:
            encode_data = event
        else:
            encode_data = {}
            for source_field in self.source_fields:
                if source_field not in event:
                    continue
                encode_data.update({source_field: event[source_field]})
                if self.drop_original:
                    event.pop(source_field, None)
        try:
            encode_data = json.dumps(encode_data)
        except:
            etype, evalue, etb = sys.exc_info()
            self.logger.warning("Could not json encode event data: %s. Exception: %s, Error: %s." % (event, etype, evalue))
            yield event
            return
        event.update({self.target_field: encode_data})
        yield event



# -*- coding: utf-8 -*-
import os
import socket
import sys

import lumbermill.utils.DictUtils as DictUtils
from lumbermill.BaseThreadedModule import BaseThreadedModule
from lumbermill.utils.Decorators import ModuleDocstringParser


@ModuleDocstringParser
class StdIn(BaseThreadedModule):
    """
    Reads data from stdin and sends it to its output queues.

    Configuration template:

    - StdIn:
       multiline:                       # <default: False; type: boolean; is: optional>
       stream_end_signal:               # <default: False; type: boolean||string; is: optional>
       receivers:
        - NextModule
    """

    module_type = "input"
    """Set module type"""
    can_run_forked = False

    def configure(self, configuration):
        BaseThreadedModule.configure(self, configuration)
        self.multiline = self.getConfigurationValue('multiline')
        self.stream_end_signal = self.getConfigurationValue('stream_end_signal')

    def run(self):
        if not self.receivers:
            self.logger.error("Will not start module %s since no receivers are set." % (self.__class__.__name__))
            return
        self.pid = os.getpid()
        hostname = socket.gethostname()
        multiline_data = ""
        while self.alive:
            data = sys.stdin.readline()
            if data.__len__() > 0:
                if not self.multiline:
                    self.sendEvent(DictUtils.getDefaultEventDict({"received_from": 'stdin://%s' % hostname, "data": data}, caller_class_name=self.__class__.__name__))
                else:
                    if self.stream_end_signal and self.stream_end_signal == data:
                        self.sendEvent(DictUtils.getDefaultEventDict({"received_from": 'stdin://%s' % hostname, "data": multiline_data}, caller_class_name=self.__class__.__name__))
                        multiline_data = ""
                        continue
                    multiline_data += data
            else: # an empty line means stdin has been closed
                if multiline_data.__len__() > 0:
                    self.sendEvent(DictUtils.getDefaultEventDict({"received_from": 'stdin://%s' % hostname, "data": multiline_data}, caller_class_name=self.__class__.__name__))
                self.lumbermill.shutDown()
                self.alive = False


# -*- coding: utf-8 -*-
import itertools
import sys

from lumbermill.BaseThreadedModule import BaseThreadedModule
from lumbermill.utils.Decorators import ModuleDocstringParser


@ModuleDocstringParser
class Permutate(BaseThreadedModule):
    """
    Creates successive len('target_fields') length permutations of elements in 'source_field'.

    To add some context data to each emitted event 'context_data_field' can specify a field
    containing a dictionary with the values of 'source_field' as keys.

    Configuration template:

    - Permutate:
       source_field:                    # <type: string; is: required>
       target_fields:                   # <type: list; is: required>
       context_data_field:              # <default: ""; type:string; is: optional>
       context_target_mapping:          # <default: {}; type: dict; is: optional if context_data_field == "" else required>
       receivers:
        - NextModule
    """

    module_type = "modifier"
    """Set module type"""

    def handleEvent(self, event):
        """
        Process the event.

        @param event: dictionary
        @return data: dictionary
        """

        try:
            context_data = event[self.getConfigurationValue('context_data_field')]
        except KeyError:
            context_data = False
        try:
            permutation_data = event[self.getConfigurationValue('source_field')]
        except KeyError:
            yield event
            return
        if type(permutation_data) is not list:
            yield event
            return
        target_field_names = self.getConfigurationValue('target_fields')
        context_target_mapping = self.getConfigurationValue('context_target_mapping')
        for permutation in itertools.permutations(permutation_data, r=len(target_field_names)):
            event_copy = event.copy()
            if context_data:
                try:
                    # Rewrite the context data keys to new keys in context_target_mapping
                    ctx_data = {}
                    for idx, dct in enumerate([context_data[key] for key in permutation if key in context_data]):
                        for mapping_key, newkeys in context_target_mapping.items():
                            if mapping_key in dct:
                               ctx_data[newkeys[idx]] = dct[mapping_key]
                    event_copy.update(ctx_data)
                except:
                    etype, evalue, etb = sys.exc_info()
                    self.logger.warning("Could not add context data. Exception: %s, Error: %s." % (etype, evalue))
            perm = dict(zip(target_field_names, permutation))
            event_copy.update(perm)
            yield event_copy


# -*- coding: utf-8 -*-
import logging
import re
import socket
import struct
import sys

import pcapy
from impacket.ImpactDecoder import EthDecoder
from impacket.ImpactPacket import IP

import lumbermill.utils.DictUtils as DictUtils
from lumbermill.BaseThreadedModule import BaseThreadedModule
from lumbermill.utils.Decorators import ModuleDocstringParser

PROTOCOL_TO_NAMES = {'eth': 'Ethernet',
                     '0x8': 'IPv4'}


@ModuleDocstringParser
class Sniffer(BaseThreadedModule):
    """
    Sniff network traffic. Needs root privileges.

    Reason for using pcapy as sniffer lib:
    As Gambolputty is intended to be run with pypy, every module should be compatible with pypy.
    Creating a raw socket in pypy is no problem but it is (up to now) not possible to bind this
    socket to a selected interface, e.g. socket.bind(('lo', 0)) will throw "error: unknown address family".
    With pcapy this problem does not exist.

    Dependencies:
     - pcapy: pypy -m pip install pcapy

    Configuration template:

    - Sniffer:
       interface:                       # <default: 'any'; type: None||string; is: optional>
       packetfilter:                    # <default: None; type: None||string; is: optional>
       promiscous:                      # <default: False; type: boolean; is: optional>
       key_value_store:                 # <default: None; type: none||string; is: optional>
       receivers:
        - NextModule
    """

    module_type = "input"
    """Set module type"""
    can_run_forked = False

    def configure(self, configuration):
        BaseThreadedModule.configure(self, configuration)
        self.interface = self.getConfigurationValue('interface')
        self.promiscous_mode = 1 if self.getConfigurationValue('promiscous') else 0
        self.kv_store = self.getConfigurationValue('key_value_store') if self.getConfigurationValue('key_value_store') else {}
        self.packet_decoders = {}
        for ether_type, ether_protocol in PROTOCOL_TO_NAMES.items():
            if "decodePacket%s" % ether_protocol in dir(self):
                self.packet_decoders[ether_type] = getattr(self, "decodePacket%s" % ether_protocol)
        try:
            self.sniffer = pcapy.open_live(self.interface, 65536, self.promiscous_mode, 100)
        except:
            etype, evalue, etb = sys.exc_info()
            self.logger.error("Sniffer could not be created. Exception: %s, Error: %s." % (etype, evalue))
            self.lumbermill.shutDown()
        if self.getConfigurationValue('packetfilter'):
            self.sniffer.setfilter(self.getConfigurationValue('packetfilter'))
        self.link_layer = self.sniffer.datalink()

    def getPacketDecoder(self, packet_protocol):
        try:
            return self.packet_decoders[packet_protocol]
        except KeyError:
            # Try to find a matching decoder.
            protocol_name = PROTOCOL_TO_NAMES[packet_protocol]
            try:
                packet_decoder_name = 'PacketDecoder%s' % protocol_name
                self.packet_decoders[packet_protocol] = globals()[packet_decoder_name]()
                return self.packet_decoders[packet_protocol]
            except:
                etype, evalue, etb = sys.exc_info()
                self.logger.error("Could not find packet decoder for protocol %s. Exception: %s, Error: %s." % (protocol_name, etype, evalue))
                return None

    def decodePacket(self, packet, packet_protocol):
        decoder = self.getPacketDecoder(packet_protocol)
        if not decoder:
            return packet
        return decoder.decodePacket(packet)

    def run(self):
        while self.alive:
            packet = None
            try:
                pcap_header, packet = self.sniffer.next()
            except:
                pass
            if not packet:
                continue
            p = self.decodePacket(packet, 'eth')
            decoded_packet = DictUtils.getDefaultEventDict({'protocols': ['ethernet'], 'data': packet, 'packet_size': pcap_header.getlen()}, caller_class_name=self.__class__.__name__)
            decoded_packet['lumbermill']['event_type'] = 'PcapSniffer'
            dest_mac, source_mac, proto_type = struct.unpack("!6s6sH", packet[:14])
            decoded_packet['source_mac'] = ':'.join('%02x' % ord(byte) for byte in source_mac)
            decoded_packet['destination_mac'] = ':'.join('%02x' % ord(byte) for byte in dest_mac)
            if proto_type == 0x8100: # 802.1Q tag (optional)
                proto_type = struct.unpack("!16xH", packet[:18])
            packet_protocol = hex(socket.ntohs(proto_type))
            decoded_packet = self.decodePacket(packet_protocol, decoded_packet)
            self.sendEvent(decoded_packet)


class BasePacketDecoder:

    def __init__(self):
        self.logger = logging.getLogger(self.__class__.__name__)
        self.packet_decoders = {}
        self.known_protocols = {}

    def getPacketDecoder(self, packet_protocol):
        if packet_protocol not in self.known_protocols:
            return None
        try:
            return self.packet_decoders[packet_protocol]
        except KeyError:
            # Try to find a matching decoder.
            protocol_name = self.known_protocols[packet_protocol]
            try:
                packet_decoder_name = 'PacketDecoder%s' % protocol_name
                self.packet_decoders[packet_protocol] = globals()[packet_decoder_name]()
                return self.packet_decoders[packet_protocol]
            except:
                #etype, evalue, etb = sys.exc_info()
                #self.logger.error("Could not find packet decoder for protocol %s. Exception: %s, Error: %s." % (protocol_name, etype, evalue))
                return None


class PacketDecoderEthernet(BasePacketDecoder):
    def __init__(self):
        BasePacketDecoder.__init__(self)
        self.decoder = EthDecoder()

    def decodePacket(self, packet):
        decoded_packet = self.decoder.decode(packet)
        next_layer_packet = decoded_packet.child()
        #pprint.pprint(dir(next_layer_packet))
        print(isinstance(next_layer_packet,IP))

class PacketDecoderIPv4(BasePacketDecoder):

    PROTOCOLS = {'0x6': 'TCP'}

    def __init__(self):
        BasePacketDecoder.__init__(self)
        self.known_protocols = PacketDecoderIPv4.PROTOCOLS

    def decodePacket(self, packet):
        """Decode an IPv4 packet. See http://en.wikipedia.org/wiki/IPv4"""
        packet_data = packet['data']
        ip_header = struct.unpack('!BBHHHBBH4s4s' , packet_data[0:20])
        version_ihl = ip_header[0]
        version = version_ihl >> 4
        ihl = version_ihl & 0xF
        ip_header_length = ihl * 4
        ttl = ip_header[5]
        ip_protocol = hex(ip_header[6])
        s_addr = socket.inet_ntoa(ip_header[8])
        d_addr = socket.inet_ntoa(ip_header[9])
        packet['protocols'].append('IPv4')
        packet.update({'version': version,
                       'ttl': ttl,
                       'source': s_addr,
                       'destination': d_addr,
                       'data': packet_data[ip_header_length:]})
        decoder = self.getPacketDecoder(ip_protocol)
        if decoder:
            packet = decoder.decodePacket(packet)
        return packet

class PacketDecoderTCP(BasePacketDecoder):

    PROTOCOLS = {
        1: 'TCPMUX',
        5: 'RJE',
        7: 'ECHO',
        9: 'DISCARD',
        11: 'SYSTAT',
        13: 'DAYTIME',
        17: 'QOTD',
        18: 'MSP',
        19: 'CHARGEN',
        20: 'FTP-DATA',
        21: 'FTP',
        22: 'SSH',
        23: 'TELNET',
        24: 'LMTP',
        25: 'SMTP',
        37: 'TIME',
        39: 'RLP',
        42: 'NAMESERVER',
        43: 'NICNAME',
        49: 'TACACS',
        50: 'RE-MAIL-CK',
        53: 'DOMAIN',
        63: 'WHOIS++',
        67: 'BOOTPS',
        68: 'BOOTPC',
        69: 'TFTP',
        70: 'GOPHER',
        71: 'NETRJS-1',
        72: 'NETRJS-2',
        73: 'NETRJS-3',
        74: 'NETRJS-4',
        79: 'FINGER',
        80: 'HTTP',
        88: 'KERBEROS',
        95: 'SUPDUP',
        101: 'HOSTNAME',
        102: 'ISO-TSAP',
        105: 'CSNET-NS',
        107: 'RTELNET',
        109: 'POP2',
        110: 'POP3',
        111: 'SUNRPC',
        113: 'AUTH',
        115: 'SFTP',
        117: 'UUCP-PATH',
        119: 'NNTP',
        123: 'NTP',
        137: 'NETBIOS-NS',
        138: 'NETBIOS-DGM',
        139: 'NETBIOS-SSN',
        143: 'IMAP',
        161: 'SNMP',
        162: 'SNMPTRAP',
        163: 'CMIP-MAN',
        164: 'CMIP-AGENT',
        174: 'MAILQ',
        177: 'XDMCP',
        178: 'NEXTSTEP',
        179: 'BGP',
        191: 'PROSPERO',
        194: 'IRC',
        199: 'SMUX',
        201: 'AT-RTMP',
        202: 'AT-NBP',
        204: 'AT-ECHO',
        206: 'AT-ZIS',
        209: 'QMTP',
        210: 'Z39.50',
        213: 'IPX',
        220: 'IMAP3',
        245: 'LINK',
        347: 'FATSERV',
        363: 'RSVP_TUNNEL',
        366: 'ODMR',
        369: 'RPC2PORTMAP',
        370: 'CODAAUTH2',
        372: 'ULISTPROC',
        389: 'LDAP',
        400: 'OSB-SD',
        427: 'SVRLOC',
        434: 'MOBILEIP-AGENT',
        435: 'MOBILIP-MN',
        443: 'HTTPS',
        444: 'SNPP',
        445: 'MICROSOFT-DS',
        464: 'KPASSWD',
        468: 'PHOTURIS',
        487: 'SAFT',
        488: 'GSS-HTTP',
        496: 'PIM-RP-DISC',
        500: 'ISAKMP',
        538: 'GDOMAP',
        535: 'IIOP',
        546: 'DHCPV6-CLIENT',
        547: 'DHCPV6-SERVER',
        554: 'RTSP',
        563: 'NNTPS',
        565: 'WHOAMI',
        587: 'SUBMISSION',
        610: 'NPMP-LOCAL',
        611: 'NPMP-GUI',
        612: 'HMMP-IND',
        631: 'IPP',
        636: 'LDAPS',
        674: 'ACAP',
        694: 'HA-CLUSTER',
        749: 'KERBEROS-ADM',
        750: 'KERBEROS-IV',
        765: 'WEBSTER',
        767: 'PHONEBOOK',
        873: 'RSYNC',
        875: 'RQUOTAD',
        992: 'TELNETS',
        993: 'IMAPS',
        994: 'IRCS',
        995: 'POP3S',
        512: 'EXEC',
        513: 'LOGIN',
        514: 'SHELL',
        515: 'PRINTER',
        519: 'UTIME',
        520: 'EFS',
        521: 'RIPNG',
        525: 'TIMED',
        526: 'TEMPO',
        530: 'COURIER',
        531: 'CONFERENCE',
        532: 'NETNEWS',
        540: 'UUCP',
        543: 'KLOGIN',
        544: 'KSHELL',
        548: 'AFPOVERTCP',
        1080: 'SOCKS'
    }

    def __init__(self):
        BasePacketDecoder.__init__(self)
        self.known_protocols = PacketDecoderTCP.PROTOCOLS
        self.ctrl_flags = {'fin': 1, 'syn': 2, 'rst': 4, 'psh': 8, 'ack': 16, 'urg': 32}
        self.flows = {}

    def decodePacket(self, packet):
        packet_data = packet['data']
        tcp_header = packet_data[0:20]
        tcph = struct.unpack('!HHLLHHHH' , tcp_header) #'!HHLLBBHHH'
        source_port = tcph[0]
        dest_port = tcph[1]
        sequence = tcph[2]
        acknowledgement = tcph[3]
        data_offset_reserved = tcph[4]
        data_offset = data_offset_reserved >> 12
        data_offset = data_offset * 4
        packet_ctrl_flags = data_offset_reserved & 0x3f
        packet['tcp_ctrl_flags'] = []
        for tcp_flag_name, tcp_flag_mask in self.ctrl_flags.items():
            if packet_ctrl_flags & tcp_flag_mask == tcp_flag_mask:
               packet['tcp_ctrl_flags'].append(tcp_flag_name)
        packet['protocols'].append('TCP')
        packet.update({'source_port': source_port,
                       'destination_port': dest_port,
                       'sequence_number': sequence,
                       'ack': acknowledgement,
                       'data': packet_data[data_offset:]})
        # Set flow id.
        flow_id = "%s %s" % tuple(sorted([ "%(source)s:%(source_port)s" % packet, "%(destination)s:%(destination_port)s" % packet ]))
        packet['flow_id'] = flow_id
        # Get protocol.
        tcp_protocol = None
        decoder = None
        if flow_id in self.flows:
            tcp_protocol = self.flows[flow_id]
        if packet['tcp_ctrl_flags'] in (['syn'], ['ack']) and flow_id not in self.flows:
            tcp_protocol = self.flows[flow_id] = dest_port
        elif packet['tcp_ctrl_flags'] in (['fin'], ['rst']) and flow_id in self.flows:
            del self.flows[flow_id]
        if tcp_protocol:
            decoder = self.getPacketDecoder(tcp_protocol)
        if decoder:
            packet = decoder.decodePacket(packet)
        # If no decoder was found at least set protocol name.
        elif tcp_protocol in self.known_protocols:
            packet['protocols'].append(self.known_protocols[tcp_protocol])
        return packet

class PacketDecoderHTTP(BasePacketDecoder):

    def __init__(self):
        BasePacketDecoder.__init__(self)
        self.re_request = re.compile(r"^(?P<http_method>\w+?)\s+(?P<url>.*?)\s+HTTP/[\d\.]+")
        self.re_headers = re.compile(r"^(?P<http_header>.*?): (?P<value>.*?)\r\n", re.MULTILINE)

    def decodePacket(self, packet):
        packet['protocols'].append('html')
        rawrequest = self.re_request.search(packet['data'])
        if rawrequest:
            packet.update(rawrequest.groupdict())
        headers = self.re_headers.findall(packet['data'])
        packet.update({header: value for header, value in headers})
        return packet